(* ::Package:: *)

(* Theorema 
    Copyright (C) 2010 The Theorema Group

    This file is part of Theorema 2.0
    
    Theorema 2.0 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Theorema 2.0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

BeginPackage[ "Theorema`Language`HORewriting`"]

Needs[ "Theorema`Language`"]
Needs[ "Theorema`Common`"]

(* Exported symbols added here with SymbolName::usage *)

With[ {lang = "English"},
	MessageName[ formulaToHORules, "usage", lang] =
		"formulaToHORules[ fml] turns 'fml' into a list of (possibly) higher-order rewrite rules. The result depends on the form of the optional parameter 'Rules', but it is essentially a combination of\n
		- forward rules,\n
		- backward rules,\n
		- substitution rules,\n
		- explicit definitions,\n
		- lazy definitions (see option 'LazySymbols',\n
		- implicit definitions.\n
		Depending on 'fml', some of these rule-lists might of course be empty.
		The individual rules are not just plain Mathematica transformation rules, but rather triples '{key, rule, ground}', where 'key' is the key of the formula the rule originated from, 'rule' is the actual transformation rule (for implicit definitions,
		it is a pair where the first element is the pattern of the rule), and 'ground' is 'True' or 'False' depending on whether the left-hand-side of 'rule' contains free variables ('False') or not ('True').
		Furthermore, if the optional parameter 'ReturnProof' is set to 'True', the whole output is not just a list of rules (consisting of triples), but rather a pair of the form '{forms, rules}', where 'rules' is the list of rules and
		'forms' is a list of triples of the form '{id, frm, prf}', where 'id' is a unique identifier, 'frm' is a part of the original formula giving rise to at least one rule in 'rules', and 'prf' is the proof of 'frm' from the original formula.\n
		Possible options are:\n
		ApplicationModuloBetaReduction -> \"Explicit\"|\"Implicit\"|\"All\" (default):
							With '\"Explicit\"', all applications of variables to arguments that should be tried to be matched modulo \[Beta]-reduction have to be applied with double square brackets.
							With '\"Implicit\"', all applications of a variable 'v' which is at least applied once with double square brackets are automatically tried to be matched modulo \[Beta]-reduction.
							With '\"All\"', really all applications of variables to arguments are tried to be matched modulo \[Beta]-reduction.\n
		BackwardWholeFormula -> True|False (default): With 'True', the backward rule corresponding to 'inner -> cond' is generated as well (if backward rules are generated), where 'fml' is either 'cond => inner' or just 'inner' (and 'cond' is 'True').\n
		CheckForFreeVars -> True|False (default): With 'True' all free variables in 'fml' are bound by new universal quantifiers ranging over the whole formula. With 'False' this security check is entirely omitted, possibly leading to errors if 'fml' contains free variables.\n
		DecomposeEquivalences -> True (default)|False: With 'True', equivalences are not only turned into substitutions, but their LHS and RHS are treated separately as well, with the other side as additional condition.\n
		LazyCheckingFunction -> d: 'd' is the name of the function which is used for checking whether a lazy definition should be expanded or not. By default, 'd' is 'None'.\n
		LazySymbols -> l: 'l' can be a list of expressions which are considered as lazy symbols in the rule generation, i.e. rules whose left-hand-side has a head which is a member of 'l' is put into the fifth output list. By default, 'l' is empty. Note that lazy-definition rules are only generated if 'ExplicitDefinition' rules are generated!\n
		ImplicitCheckingFunction -> f: 'f' is the name of the function which is used for checking conditions of implicit definitions. By default, 'f' is 'None', in which case conditions are sown with 'Sow', just as in any other rule category.\n
		NormalizeArgumentLists -> \"Lengthen\" (default)|\"Shorten\"|False: Specifies whether the argument lists of higher-order variables shall be normalized, such that all occurrences of a higher-order variable are applied to the same number of arguments (if none of the arguments is a free sequence variable).
							With \"Lengthen\", normalization happens through \[Eta]-expansion (i.e. introducing additional \[Lambda]-abstractions) and un-Currying (i.e. transforming 'F[ a][ b]' into 'F[ a, b]').
							With \"Shorten\", normalization happens through Currying (i.e. transforming 'F[ a, b]' into 'F[ a][ b]').
							With 'False' (or any value different from \"Lengthen\" and \"Shorten\"), no normalization happens at all and the formula is left unchanged; higher-order variables that are not always applied to the same number of arguments are treated as ordinary first-order variables.
							Note that \"Lengthen\" and \"Shorten\" only affect the LHSsof rewrite-rules: once the arity of a higher-order variable has been determined by lengthing/shortening argument lists on the LHS, all argument lists of on the RHS are normalized to precisely that arity by lengthening AND shortening. With 'False', the entire formula/rule is left completely unchanged.\n
		RandomRanges -> True (default)|False: With 'True', simple ranges of quantifiers can match arbitrary (random) ranges in an expression the rules are applied on, in some cases.\n
		RenameBoundVars -> True (default)|False: With 'True', bound variables in 'fml' are renamed to have unique variable names. This is important for the rule generation, and hence this option must only be set to 'False' if the variables already have to correct names and shapes!\n
		ReturnProof -> True|False (default): Specifies whether the part of the original formula exactly corresponding to the generated rule, together with its proof, shall be returned as well. 'ReturnProof' only takes effect if 'Type' is \"Proof\".\n
		RightToLeft -> True (default)|False: With 'True', equalities and equivalences are turned into rules both from left-to-right and from right-to-left.\n
		Rules -> r: 'r' can be any combination of '\"Forward\"', '\"Backward\"', '\"Substitution\"', '\"ExplicitDefinition\"', '\"ImplicitDefinition\"', '\"LazyDefinition\"' and '\"All\"' (default; '\"All\"' is a mere shortcut for the 6-tuple '{\"Forward\", \"Backward\", \"Substitution\", \"ExplicitDefinition\", \"LazyDefinition\", \"ImplicitDefinition\"}'). The output of the function follows precisely the shape of 'Rules'.\n
		RuleOutput -> expr: 'expr' can be an arbitrary expression containing any of the four tokens \"NewExpression\", \"Condition\", \"Instantiation\" and \"Formula\", specified the output of the generated transformation rules. Each of the tokens is replaced by its concrete value, of course:
							\"NewExpression\" by the instance of the RHS of the rule,
							\"Condition\" by the additional condition,
							\"Instantiation\" by the instantiations of the universally quantified variables (Instantiations are represented as a list of expressions the universal quantifiers have to be instantiated with from left to right. Note that some of the expressions might contain '$TMHOR$Unknown'.),
							\"Formula\" by the original formula the rule originates from, and
							\"Key\" by the key of the original formula, and
							\"ReferenceID\" by the unique identifier of the part of the original formula the rule corresponds to; this value is -1 if option 'ReturnProof' is 'False'.
							Note that 'RuleOutput' only has an effect if 'type' is \"Proof\"; in computations, rules *always* only return the new expression (and check additional conditions if necessary).\n
		RuleSow -> l: 'l' can be any list of elements of the form '{key -> value}', where the 'key' are arbitrary strings and the 'value' are arbitrary expressions, possibly containing the five tokens of 'RuleOutput'. If 'type' is \"Proof\", then every generated rule, when applied, sows each of the 'value' with tag 'key' (with the tokens replaced by their concrete values, of course).\n
		Strict -> True (default)|False: With 'True', every higher-order variable must be applied at least once on an argument involving a local variable, in order to be treated as a higher-order variable.\n
		Type -> \"Proof\" (default)|\"Computation\": Secifies whether Mathematica transformation rules suitable for rewriting in proofs ('RuleDelayed') or definitions suitable for computations ('SetDelayed') shall be generated. Value \"Computation\" is experimental and does not work properly yet!";
		
	MessageName[ hoRewrite, "usage", lang] =
		"hoRewrite[ expr, rules] applies the rules in 'rules' to 'expr', taking care of variable renaming (for having unique names).\n
		Option 'Method -> m' can be used to specify the method which shall be used for applying the rules (e.g. any of the 'replace...AndTrack'). The default is 'replaceAllAndTrack'.";
		
	MessageName[ hoInstantiate, "usage", lang] = "hoInstantiate[ form, inst] instantiates formula 'form' by 'inst', where 'inst' is a list of instantiations as sowed when applying a higher-order rewrite rule. Meta-level $TMHOR$Lambda-binders in instances of higher-order variables are immediately \[Beta]-reduced, if possible; all other redexes are kept.";
	
	MessageName[ preprocessExpression, "usage", lang] = "preprocessExpression[ expr] prepares 'expr' for being rewritten by higher-order rewrite rules. This function is automatically called if 'hoRewrite' is used; Otherwise, it must be called manually.";
	
	MessageName[ postprocessExpression, "usage", lang] = "postprocessExpression[ expr] transformes 'expr' back to a standard Theorema expression. This function is automatically called if 'hoRewrite' is used; Otherwise, it must be called manually.";
	
	MessageName[ generateGroundRule, "usage", lang] = "generateGroundRule[ l, r, c] generates a single rewrite-rule, where expressions alpha-equivalent to 'l' are replaced by 'r' if all conditions in the list 'c' hold. 'l', 'r', and 'c' must not contain free variables.";
]

Options[ hoRewrite] = {Method -> Theorema`Common`replaceAllAndTrack};
Options[ hoInstantiate] = {EliminateNonSimpleRanges -> True, CurryUncurry -> True, BetaReduce -> True};
Options[ formulaToHORules] = {
		Rules -> "All",
		ApplicationModuloBetaReduction -> "All",
		RandomRanges -> True,
		RightToLeft -> True,
		RenameBoundVars -> True,
		CheckForFreeVars -> False,
		DecomposeEquivalences -> True,
		Strict -> True,
		LazySymbols -> {},
		ImplicitCheckingFunction -> None,
		NormalizeArgumentLists -> "Lengthen",
		LazyCheckingFunction -> None,
		BackwardWholeFormula -> False,
		Type -> "Proof",
		RuleSow -> {"cond" -> "Condition", "ref" -> "Formula"},
		RuleOutput -> "NewExpression",
		ReturnProof -> False
	}

Begin[ "`Private`"] (* Begin Private Context *)



(* ::Section::Closed:: *)
(* General-Purpose Functions *)

simplifiedNotProof[ Not$TM[ Not$TM[ A_]], dir_] :=
	{
		A,
		Switch[ dir,
			Forward,
			Hold[ "CContr0", {"Contr", 1}],
			_,
			Hold[ "NotI", {"Contr", A}]
		]
	}
simplifiedNotProof[ form:(Not$TM[ And$TM[ A__]]), dir_] :=
	{
		Map[ Not$TM, Or$TM[ A]],
		Switch[ dir,
			Forward,
			Hold[ {"NotConjD", form}, "Triv"],
			_,
			Hold[ "NotConjI", "Triv"]
		]
	}
simplifiedNotProof[ form:(Not$TM[ Or$TM[ A__]]), dir_] :=
	{
		Map[ Not$TM, And$TM[ A]],
		Switch[ dir,
			Forward,
			Hold[ {"NotDisjD", form}, "Triv"],
			_,
			Hold[ "NotDisjI", "Triv"]
		]
	}
simplifiedNotProof[ form:(Not$TM[ Implies$TM[ A_, B_]]), dir_] :=
	{
		And$TM[ A, Not$TM[ B]],
		Switch[ dir,
			Forward,
			Hold[ {"NotImpD", form}, "Triv"],
			_,
			Hold[ "NotImpI", "Triv"]
		]
	}
simplifiedNotProof[ form:(Not$TM[ Iff$TM[ A_, B_]]), dir_] := 
	{
		Or$TM[ And$TM[ A, Not$TM[ B]], And$TM[ Not$TM[ A], B]],
		Switch[ dir,
			Forward,
			Hold[ {"NotEquivD", form}, "Triv"],
			_,
			Hold[ "NotEquivI", "Triv"]
		]
	}
simplifiedNotProof[ form:(Not$TM[ True]), dir_] := 
	{
		False,
		Switch[ dir,
			Forward,
			Hold[ {"NotE", form}, "Triv"],
			_,
			Hold[ "Contr"]
		]
	}
simplifiedNotProof[ Not$TM[ False], dir_] := 
	{
		True,
		Switch[ dir,
			Forward,
			Hold[ "Triv"],
			_,
			Hold[ "NotI", "Contr"]
		]
	}
simplifiedNotProof[ A_, _] :=
	{A, Hold[ "Triv"]}


(* ::Section::Closed:: *)
(* Auxiliary Functions *)


makeUniqueVarNames[ expr_] :=
	With[ {repl = Join @@ Cases[
						analyzeVars[ expr],
						(VAR$[ v:Except[ _List]] -> {fp_, bp_}) :>
						With[ {tmp = Replace[ bp, (p_ -> pl_) :> (Prepend[ pl, p] -> VAR$[ {$varIndex++, v}]), {1}]},
							If[ fp === {},
								tmp,
								Prepend[ tmp, fp -> VAR$[ {$varIndex++, v}]]
							]
						]
					]
		},
		ReplacePart[ expr, repl]
	]

myTranspose[ h_[ ]] := h[ h[ ], h[ ]]
myTranspose[ expr_] := Transpose[ expr]

replaceAllBound[ v_VAR$, new_, vars_List] := If[ MemberQ[ vars, v], new, v]
replaceAllBound[ q_[ r_RNG$, rest___], new_, vars_List] :=
	With[ {newVars = Join[ rngVariables[ r], vars]},
		Apply[ replaceAllBound[ q, new, vars], Map[ replaceAllBound[ #, new, newVars]&, {r, rest}]]
	]
replaceAllBound[ h_[ args___], new_, vars_List] :=
	Apply[ replaceAllBound[ h, new, vars], Map[ replaceAllBound[ #, new, vars]&, {args}]]
replaceAllBound[ expr_, _, _List] := expr


(* ::Section::Closed:: *)
(* Rule Application *)


hoRewrite[ expr_, rules_List, opts___?OptionQ] :=
	With[ {repl = Method /. {opts} /. Options[ hoRewrite], pp = preprocessExpression[ expr]},
		postprocessExpression[ repl[ pp, rules]]
	]
hoRewrite[ expr_, rule:(_Rule|_RuleDelayed), opts___] := hoRewrite[ expr, {rule}, opts]


(* ::Subsection:: *)
(* Trusted Rewriting *)

hoInstantiate[ FML$[ _, form_, __], inst_List, opts___?OptionQ] := hoInstantiate[ form, inst, opts]
hoInstantiate[ form_, {}, ___?OptionQ] :=
	form
hoInstantiate[ expr:((q:Forall$TM|Exists$TM)[ rng_RNG$, c_, _]), inst_List, opts___?OptionQ] /; Length[ rng] === Length[ inst] :=
	With[ {conn = Switch[ q, Forall$TM, Implies$TM, _, And$TM]},
	Module[ {br, cu, ensr, newInst, newE, rc, cond, form},
		{br, cu, ensr} = {BetaReduce, CurryUncurry, EliminateNonSimpleRanges} /. {opts} /. Options[ hoInstantiate];
		
		If[
			(* Check whether instances match type of variables (individual vs. sequence), and adjust 'inst' by replacing 'SEQ$[ a]' by just 'a'. *)
			Catch[
				newInst =
					MapThread[
						Function[ {t, e},
							newE = e;
							While[ MatchQ[ newE, SEQ$[ _]], newE = First[ newE]];
							Switch[ t,
								Symbol,
								If[ isIndividual[ newE],
									newE,
									Throw[ False]
								],
								
								SEQ1$,
								If[ isNonEmptySequence[ newE],
									newE,
									Throw[ False]
								],
								
								_,
								newE
							]
						],
						{sequenceType /@ rngVariables[ rng], inst}
					];
				Throw[ True]
			],
			
			{rc, cond, form} = renameBound[ List @@ expr, newInst, "checkTypes" -> False, "postprocessing" -> Identity];
			rc = rngToCondition[ rc];
			
			form = Switch[ c, True, form, _, conn[ cond, form]];
			(* The decision whether to explicitly add 'cond' or not is based on the original input 'c', in order to be able to tell in advance what is going to happen. *)
			
			Switch[ rc,
				{_},
				form = conn[ First[ rc], form],
				{__},
				form = conn[ And$TM @@ rc, form]
			];
			
			form = sequenceFlatten[ form];
			
			If[ TrueQ[ br], form = betaReduce[ form, cu, ensr]];
		
			(* Post-processing: turn some $TMHOR$Lambda into Lambda$TM *)
			form //. $TMHOR$Lambda[ r_, inner_] :> Switch[ r, RNG$[ ___SIMPRNG$], Lambda$TM, _, $TMHOR$Lambda][ r, True, inner],
			
			$Failed
		]
		
		(*
		instRanges = Replace[ instRanges, all:{{pre___, _} -> _, ___Rule} :> ({pre} -> Replace[all, ({___, l_} -> val_) :> (l -> val), {1}]), {1}];
		instRanges = SortBy[ instRanges, Composition[ Length, First]];
		Scan[ Function[ ir,
				With[ {pos = First[ ir]},
					qu = First[ Extract[ out, {pos}]];
					If[ MatchQ[ qu, Forall$TM[ _RNG$, _, _]],
						instRng = First[ qu];
						rngVars = rngVariables[ instRng];
						repl = Select[ Last[ ir], (1 <= First[ #] <= Length[ instRng])&];
						instRng = Part[ instRng, repl[[All, 1]]];
						qu[[1]] = Select[ First[ qu], (!MemberQ[ instRng, #])&];
						AppendTo[ qu, rngToCondition[ instRng]];	(* We temporarily append the new range-condition to the universal quantifier, only not to mess up indices. *)
						qu = substituteFree[ qu, Replace[ repl, (i_ -> expr_) :> (rngVars[[i]] -> expr), {1}]];
						out = If[ pos === {}, qu, ReplacePart[ out, pos -> qu]]
					]
				]
			],
			instRanges
		];
		
		out = out //. {
						Forall$TM[ RNG$[], cond_, inner_, conds_List] :>
							With[ {newCond = makeConjunction[ DeleteCases[ Prepend[ conds, cond], True], And$TM]},
								If[ newCond === True,
									inner,
									Implies$TM[ newCond, inner]
								]
							],
						Forall$TM[ r:RNG$[ __], cond_, inner_, conds_List] :>
							Forall$TM[ r, makeConjunction[ DeleteCases[ Prepend[ conds, cond], True], And$TM], inner],
						Part$TM[ l_$TMHOR$Lambda, args___] :> l[ args],
						(seq:SEQ$[ ___])[ args___] :> Map[ (#[ args])&, seq]
					}
				/. SEQ$ -> Sequence;
				
		If[ TrueQ[ br], out = betaReduce[ out, cu, ensr]];
		
		(* post-processing: turn some $TMHOR$Lambda into Lambda$TM *)
		out //. $TMHOR$Lambda[ r_, inner_] :> Switch[ r, RNG$[ ___SIMPRNG$], Lambda$TM, _, $TMHOR$Lambda][ r, True, inner]
		*)
	]
	]
hoInstantiate[ ___] := $Failed
	
(* 'betaReduce[ expr]' reduces all beta-redexes of $TMHOR$Lambda expressions, as far as possible.
The optional argument 'curry' specifies whether
- expressions of the form '$TMHOR$Lambda[ ...][ args1][ args2]' shall be uncurried to
	'$TMHOR$Lambda[ ...][ args1, args2]' if necessary, i.e. if there are more bound variables than
	arguments in 'args1', and
- expressions of the form '$TMHOR$Lambda[ ...][ args1, args2]' shall be curried to
	'$TMHOR$Lambda[ ...][ args1][ args2]' if necessary, i.e. if there are fewer bound variables
	than arguments in 'args1, args2'.
In any case, '$TMHORLambda[ RNG$[ r1, r2], inner]' is split into
'$TMHOR$Lambda[ RNG$[ r1], $TMHOR$Lambda[ RNG$[ r2], inner]]' whenever there are too many bound
variables; the converse operation is never applied, though.
Note that an expression involving a non-simple range, say '$TMHOR$Lambda[ SETRNG$[ x, A], P[ x]]',
is equivalent to, say, '$TMHOR$Lambda[ x, Element[ x, A] && P[ x]]'; this does *NOT* mean that a
redex can only be reduced if the respective argument is known to belong to the range! This is most
probably different to the intended semantics of non-simple ranges in the object-level Lambda
binder. Any $TMHOR$Lambda remaining in the output (now again with a third argument) signals that
something went wrong. *)
(* TODO: Ensure to only beta-reduce terms that yield individuals. *)
betaReduce[ expr_, curry_, elimNonSimple_] :=
	Module[ {out = expr},
		If[ TrueQ[ curry],
			out = out //. $TMHOR$Lambda[ r_RNG$, inner_][ a1___][ a2___] :>
							With[ {lr = Length[ r], l1 = Length[ Hold[ a1]]},
								$TMHOR$Lambda[ r, inner][ a1, a2] /; l1 < lr
							]
		];
		
		(* normalize the number of arguments of each $TMHOR$Lambda, such that it agrees with the number of bound variables *)
		out =
			out //. $TMHOR$Lambda[ r_RNG$, inner_][ args___] :>
					With[ {lr = Length[ r], la = Length[ Hold[ args]]},
						If[ lr < la,
							If[ lr === 0,
								inner[ args],
								($TMHOR$Lambda[ r, inner] @@ Take[ Hold[ args], lr]) @@ Drop[ Hold[ args], lr]
							],
							If[ la === 0,
								$TMHOR$Lambda[ r, inner],
								$TMHOR$Lambda[ Take[ r, la], $TMHOR$Lambda[ Drop[ r, la], inner]][ args]
							]
						] /; ((la < lr && (la =!= 0 || TrueQ[ curry])) || (lr < la && (lr === 0 || TrueQ[ curry])))
					];
					
		If[ TrueQ[ elimNonSimple],
			(* eliminate all non-simple ranges *)
			out =
				out //. q_[ rng_RNG$, cond_, rest__] :>
						Module[ {newRng, newCond},
							q[ newRng, newCond, rest] /;
								(MemberQ[ cond, $TMHOR$Lambda[ RNG$[ ___, Except[ _SIMPRNG$], ___], _], {0, Infinity}, Heads -> True] &&
								({newRng, newCond} = eliminateNonSimpleRanges[ rng, cond]) =!= {rng, cond})
						]
					//. all:($TMHOR$Lambda[ RNG$[ ___, Except[ _SIMPRNG$], ___], _]) :> eliminateNonSimpleRanges[ all]
		];
				
		(* beta-reduce *)
		out	//. $TMHOR$Lambda[ r:RNG$[ ___SIMPRNG$], inner_][ args___] /; Length[ r] === Length[ Hold[ args]] :>
					With[ {res = substituteFree[ inner, Thread[ rngVariables[ r] -> {args}]]}, res /; res =!= $Failed]
	]
	
eliminateNonSimpleRanges[ rng_RNG$, expr_] :=
	eliminateNonSimpleRanges[ rng, expr, False]
eliminateNonSimpleRanges[ rng_RNG$, conds_And$TM, False] :=
	Module[ {newRng = rng, c0, out},
		out = Map[
			Function[ c,
				{newRng, c0} = eliminateNonSimpleRanges[ newRng, c, False];
				c0
			],
			conds
		];
		{newRng, out}
	]
eliminateNonSimpleRanges[ rng1_RNG$, $TMHOR$Lambda[ rng2_RNG$, inner_][ args___], True|False] :=
	Module[ {newRng1, newRng2, newCond},
		{newRng1, newRng2} = eliminateNonSimpleRanges[ rng1, rng2, {args}];
		newCond = $TMHOR$Lambda[ newRng2, inner][ args];
		If[ MemberQ[ newRng2, Except[ _SIMPRNG$]],
			newCond = eliminateNonSimpleRanges[ newCond]
		];
		{newRng1, newCond}
	]
(* It could be that the head of the head of the ... of 'h' is $TMHOR$Lambda ... *)
eliminateNonSimpleRanges[ rng_RNG$, h_[ args___], True|False] :=
	Module[ {newRng, new},
		{newRng, new} = eliminateNonSimpleRanges[ rng, h, True];
		{newRng, new[ args]}
	]
eliminateNonSimpleRanges[ rng_RNG$, expr_, True|False] :=
	{rng, expr}
eliminateNonSimpleRanges[ rng1_, rng2:RNG$[ ___, Except[ _SIMPRNG$], ___], args_List] /; Length[ rng2] === Length[ args] :=
	Module[ {newRng1 = rng1, newRng2, rv1 = Reverse[ rngVariables[ rng1]], rv2 = Reverse[ rngVariables[ rng2]], p, rc},
		newRng2 =
			RNG$ @@ MapThread[
				Function[ {vr2, a},
					If[ MatchQ[ vr2, _SIMPRNG$] || !MatchQ[ a, _VAR$],
						vr2,
						
						p = Position[ rv1, a, {1}, 1];
						If[ p === {} || !MatchQ[ newRng1[[Length[ rv1] + 1 - p[[1, 1]]]], _SIMPRNG$],
							vr2,
							
							p = p[[1, 1]];
							rc = Replace[
										freeVariables[ Rest[ vr2]],
										v_ :>
											With[ {u = Position[ rv2, v, {1}, 1]},
												If[ u === {},
													Null,
													v -> args[[Length[ args] + 1 - u[[1, 1]]]]
												]
											],
										{1}
									];
							rc = substituteFree[ Rest[ vr2], rc, "checkTypes" -> False, "postprocessing" -> Identity];
							If[ Intersection[ Take[ rv1, p - 1], freeVariables[ rc]] =!= {},
								(* If a variable occurring free in 'rc' is bound only *after* 'a' in 'rng1', the range cannot be moved. *)
								vr2,
								
								newRng1 = ReplacePart[ newRng1, Length[ rv1] + 1 - p -> Prepend[ rc, a]];
								SIMPRNG$[ First[ vr2]]
							]
						]
					]
				],
				{List @@ rng2, args}
			];
		{newRng1, newRng2}
	]
eliminateNonSimpleRanges[ rng1_, rng2:RNG$[ ___SIMPRNG$], _List] :=
	{rng1, rng2}
eliminateNonSimpleRanges[ all:($TMHOR$Lambda[ rng_RNG$, inner_])] :=
	Module[ {newRng, newInner, conds = {}, rename = {}, rename0, i = 0, varStruct = analyzeVars[ all]},
		Scan[
			Function[ vr,
				++i;
				If[ !MatchQ[ vr, _SIMPRNG$],
					(* All variables that appear free in 'Rest[ vr]' and are bound afterward in 'rng' have to be renamed,
					in order to avoid variable capture when adjoining the range-condition to 'inner'. *)
					rename = Join[ rename,
								Cases[ varStruct,
									(v_ -> {{___, {1, i, _?(# > 1 &), ___}, ___}, bp_}) :>
										(v -> Flatten[ Cases[ bp, ({1, j_, 1} -> l_List) /; j > i :> Prepend[ l, {1, j, 1}]], 1])
								]
							]
				]
			],
			rng
		];
		If[ rename =!= {},
			rename0 = rename;
			rename = DeleteCases[
						Map[
							With[ {v = First[ #], l = Last[ #]},
								rename0 = Rest[ rename0];
								v -> Select[ l, (!MemberQ[ rename0, v -> {___, #, ___}])&]
							]&,
							rename
						], 
						_ -> {}
					];
			rename = MapThread[
						(Last[ #1] -> #2)&,
						{rename, freshName[ rename[[All, 1]], Theorema`Language`FormulaManipulation`Private`varName /@ Cases[ all, _VAR$, Infinity, Heads -> True]]}
					]
		];
		{newRng, newInner} = ReplacePart[ {rng, inner}, rename];
		newRng =
			Map[
				Function[ vr,
					++i;
					If[ MatchQ[ vr, _SIMPRNG$],
						vr,
						
						conds = Join[ conds, Theorema`Language`FormulaManipulation`Private`singleRngToCondition[ vr]];
						SIMPRNG$[ First[ vr]]
					]
				],
				newRng
			];
		$TMHOR$Lambda[ newRng, makeConjunction[ Append[ conds, newInner], And$TM]]
	]

(* 'preprocessExpression[ expr ]' makes the names of all variables bound in 'expr' distinct,
	by replacing 'VAR$[ x]' by 'VAR$[ {i, x}]'. *)
preprocessExpression[ expr_] :=
	($varIndex = 0; makeUniqueVarNames[ expr])

(* 'postprocessExpression[ expr]' turns all 'VAR$[ {i, x}]' in 'expr' back to 'VAR$[ x]'.
	If some 'VAR$[ {i, x}]' appears in the scope of another 'VAR$[ {j, x}]' with different
	index but same base name, 'VAR$[ {j, x}]' gets a fresh base name to avoid variable capture. *)
postprocessExpression[ expr_] :=
	Module[ {varStruct = analyzeVars[ expr], rename = {}, rn = {}, bp},
		varStruct = GatherBy[ varStruct, Composition[ Theorema`Language`FormulaManipulation`Private`varName, First]];
		Scan[
			Function[ g,
				Scan[
					Function[ map,
						With[ {x = First[ map]},
							bp = Cases[ map[[2, 2]],
										(bp0_ -> rest_List) :> Prepend[ rest, bp0] /;
											MemberQ[ g,
												_?(# =!= x &) ->
													({{___, _?(Theorema`Language`FormulaManipulation`Private`inScope[ #, bp0]&), ___}, _List}|
													{_List,
														{___, _?(!Theorema`Language`FormulaManipulation`Private`inScope[ #, bp0]&) ->
															{___, _?(Theorema`Language`FormulaManipulation`Private`inScope[ #, bp0]&), ___}, ___}
													})
											]
								];
							(* 'bp' is the list of all binding-positions of variable 'x' that need to be
								renamed in order to avoid variable capture. *)
							If[ bp =!= {},
								AppendTo[ rename, x -> Flatten[ bp, 1]];
								AppendTo[ rn, x]
							]
						]
					],
					g
				]
			],
			varStruct
		];
		rn = MapThread[ Rule, {rn, freshName[ Replace[ rn, VAR$[ {_, n_}] :> VAR$[ n], {1}], Theorema`Language`FormulaManipulation`Private`varName /@ varStruct[[All, 1, 1]]]}];
		rename =
			Replace[
				rename,
				(x_ -> all_List) :> (all -> Replace[ x, rn]),
				{1}
			];
		ReplacePart[ expr, rename] /. VAR$[ {_, n_}] :> VAR$[ n]
	]


(* ::Section::Closed:: *)
(* Rule Generation *)


generateGroundRule[ l_, r_, c_, opts___?OptionQ] :=
	Module[ {lp, rp, cp, pp},
		pp = "Preprocess" /. {opts} /. "Preprocess" -> True;
		{lp, rp, cp} = If[ pp === False,
							{l, r, c},
							preprocessExpression[ {l, r, c}]
						];
   		generateHORule[ lp, rp, cp, {}, {}, opts]
   	]

formulaToHORules[ orig:FML$[ _, form_, __], opts___?OptionQ] := formulaToHORules[ form, opts, Reference -> orig]
formulaToHORules[ form_, opts___?OptionQ] :=
	With[ {allTypes = {"Forward", "Backward", "Substitution", "ExplicitDefinition", "LazyDefinition", "ImplicitDefinition"}},
	Module[ {frm = form, types, ambr, rb, dp, stripUniv, rules, args, f, i, index, bw, out, bIndex = Null, cf, lazy, r2l, de, tried, gen, su, retProof, allRefs = {}},
		{types, ambr, rb, dp, bw, cf, lazy, r2l, de, retProof} =
			{Rules, ApplicationModuloBetaReduction, RenameBoundVars, LazySymbols, BackwardWholeFormula, CheckForFreeVars, LazySymbols, RightToLeft, DecomposeEquivalences, ReturnProof}
			/. {opts} /. Options[ formulaToHORules];
		
		If[ TrueQ[ rb],
			frm = preprocessExpression[ frm]
		];
		
		If[ cf,
			cf = Map[ (# -> VAR$[ {$varIndex++, First[ #]}])&,
					  DeleteDuplicates[ Cases[ frm, VAR$[ Except[ _List]], {0, Infinity}, Heads -> True]]
				 ];
			If[ cf =!= {},
				frm = Forall$TM[ RNG$@@Map[ SIMPRNG$[ Last[ #]]&, cf], True, frm /. cf]
			]
		];
		
		types = types /. "All" -> allTypes;
		
		(* Create a list of 4-tuples '{expr, cond, vars, prf}'. *)
		(* stripUniv = stripAllUniversalQuantifiers[ frm]; *)
		
		stripUniv = decomposeFormula[ frm, de];
		
		(* Add list 'ho' to each tuple in 'stripUniv', containing the higher-order variables.
			Modify all 'expr's and 'cond's such that application with '[[]]' is replaced by normal application. *)
		stripUniv = stripUniv /. all:{Except[ _String], _List, _List, _Hold, ___} :> getHigherOrderVars[ ambr, all];
		
		(* For each type of rules: Traverse 'stripUniv' and for each 5-tuple '{expr, cond, vars, ho, prf}' create rewrite rules.
		The rules of "first" tuple (i.e. with as many free variables as possible) where at least in one rule every higher-order
		variable can be treated as such is returned, unless this is never possible. In that case, the rewrite rules corresponding
		to the first tuple giving rise to a non-empty list of rewrite rules are returned. *)
		out = Map[
			Function[ type,
				If[ FreeQ[ types, type],
					{},
					
					tried = {};
					gen = {};
					Join @@ Map[
						Module[ {p = #},
							index = Null;
							Catch[
								While[ ListQ[ p],
									su = If[ p === {}, stripUniv, Extract[ stripUniv, p]];
									If[ MemberQ[ tried, p],
										If[ MemberQ[ gen, {a___, 0} /; MatchQ[ p, {a, __}]],
											(* continue with next leaf *)
											Throw[ {}],
											
											(* stop here *)
											p = False
										],
										
										AppendTo[ tried, p]
									];
									If[ ListQ[ p],
										args = DeleteCases[ makeHORules[ Take[ su, 5], type, lazy, r2l], $Failed];
										If[ !ListQ[ index] && args =!= {}, index = p];
										
										rules = addRules[ retProof, Hold[ allRefs], args, SignalFailure -> True, opts];
										f = Position[ rules, Except[ $Failed], {1}, Heads -> False];
										If[ f =!= {},
											AppendTo[ gen, p];
											If[ type === "Backward", bIndex = p];
											Throw[ Join[
														DeleteCases[ rules, $Failed],
														addRules[ retProof, Hold[ allRefs], Delete[ args, f], SignalFailure -> False, opts]
													]
											]
										];
										
										i = Length[ p];
										While[ i > 0 && p[[i]] === 0, --i];
										p = Take[ p, i];
										If[ p === {},
											p = False,
											p[[-1]] = 0
										]
									]
								];
								If[ ListQ[ index],
									AppendTo[ gen, index];
									If[ type === "Backward", bIndex = index];
									su = If[ index === {}, stripUniv, Extract[ stripUniv, index]];
									args = DeleteCases[ makeHORules[ Take[ su, 5], type, lazy, r2l], $Failed];
									Throw[ DeleteCases[
												addRules[ retProof, Hold[ allRefs], args, SignalFailure -> False, opts],
												$Failed
											]
									],
									Throw[ {}]
								]
							]
						]&,
						Position[ stripUniv, {_, _List, _List, _List, _Hold, _?(MatchQ[ type, #] &)}, Heads -> False]
					]
				]
			], allTypes
		];
		If[ bw && ListQ[ bIndex] && !FreeQ[ types, "Backward"],
			If[ bIndex =!= {} && bIndex =!= {0},
				su = Take[ Switch[ stripUniv, _List, stripUniv, _, Head[ stripUniv]], 5];
				args = DeleteCases[ makeHORules[ su, "Backward", lazy, r2l], $Failed];
				If[ args =!= {},
					args = {First[ args]};
					out = Insert[ out, First[ addRules[ retProof, Hold[ allRefs], args, SignalFailure -> False, opts]], {2, 1}]
				]
			]
		];
		out = types /. Thread[ allTypes -> out];
		If[ TrueQ[ retProof],
			{allRefs, out},
			out
		]
	]
	]
	
addRules[ True, Hold[ forms_], args_List, opts___] :=
	Module[ {i, id},
		Replace[
			args,
			{{a__}, {frm_, prf_}} :>
			(
				i = Position[ forms, {_, _?(alphaEquivalent[ #, frm]&), _}, {1}, 1, Heads -> False];
				If[ i === {},
					id = RandomInteger[ 1000000000];
					AppendTo[ forms, {id, frm, prf}],
					
					i = First[ First[ i]];
					id = forms[[i, 1]];
					If[ Length[ prf] < Length[ forms[[i, -1]]], forms = ReplacePart[ forms, i -> {id, frm, prf}]]
				];
				Block[ {$varIndex = $varIndex}, generateHORule[ a, ReferenceID -> id, opts]]
			),
			{1}
		]
	]
addRules[ _, _, args_List, opts___] :=
	Replace[
		args,
		{{a__}, ___} :>
			Block[ {$varIndex = $varIndex},
				generateHORule[ a, opts]
			],
		{1}
	]
	
getHigherOrderVars[ ambr_String, {expr_, cond_List, vars_List, rest___}] :=
	Module[ {newExpr = {expr, cond}, hoVars = {}},
		If[ ambr === "All",
			newExpr = newExpr //. {Part$TM[ v_, args___] :> Module[ {h = getVarHead[ v]}, (If[ !MemberQ[ hoVars, h], AppendTo[ hoVars, h]]; v[ args]) /; MemberQ[ vars, h]]};
			hoVars = DeleteDuplicates[ Join[ hoVars, Cases[ newExpr, v_VAR$[ ___] /; MemberQ[ vars, v] :> v, {0, Infinity}, Heads -> True]]],
			
			hoVars = DeleteDuplicates[ Cases[ newExpr, Part$TM[ v_VAR$, ___] /; MemberQ[ vars, v] :> v, {0, Infinity}, Heads -> True]];
			If[ ambr === "Explicit",
				hoVars = With[ {tmp = newExpr //. Part$TM[ v_VAR$, args___] :> Hold[ args]}, Select[ hoVars, FreeQ[ tmp, #]&]]
			];
			newExpr = newExpr //. {Part$TM[ v_, args___] /; MemberQ[ hoVars, getVarHead[ v]] :> v[ args]}
		];
		
		Join[ newExpr, {vars, hoVars, rest}]
	]
	
getVarHead[ x_VAR$] := x
getVarHead[ f_[ ___]] := getVarHead[ f]
getVarHead[ x_] := x
	
normalizeArities[ l_, r_, is_List, v_List, g_List, nal_] :=
	Module[ {newExpr, ho, ar, tmp, rules},
		newExpr = {l, r};
		
		Switch[ nal,
			"Lengthen"|"Shorten",
			With[ {fnc = If[ nal === "Lengthen", Max, Min]},
				ho = Select[ v,
						Function[ var,
							tmp = DeleteDuplicates[ Cases[ l, var[ args___] :> Length[ Hold[ args]], {0, Infinity}, Heads -> True]];
							If[ tmp === {},
								False,
								tmp = fnc @@ tmp;
								If[ tmp > 0,
									ar[ var] = tmp;
									True,
									False
								]
							]
						]
					]
			];
			(* We Curry / un-Curry argument lists and remove all HO vars that are applied on a global sequence variable. *)
			ho = Select[ ho,
				With[ {arity = ar[ #]},
					tmp = newExpr;
					tmp = tmp /. # -> #[] //.
							{
								#[ a1___][ a2___] :>
									Module[ {la1 = Length[ Hold[ a1]]},
										(
											If[ la1 + Length[ Hold[ a2]] <= arity,
												#[ a1, a2],
												(#@@Join[ Hold[ a1], Take[ Hold[ a2], arity - la1]])@@Drop[ Hold[ a2], arity - la1]
											]
										) /; la1 < arity
									],
								e:(#[ a__]) /; Length[ Hold[ a]] > arity :> (Take[ e, arity])@@Drop[ e, arity]
							};
					If[ FreeQ[ tmp, #[ ___, _?isFreeSeqTerm[ g], ___]],
						newExpr = tmp;
						True,
						False
					]
				]&
			];
			newExpr = newExpr //.
						Map[
							With[ {arity = ar[ #]},
								e:(#[ a___]) :> Module[ {la = Length[ Hold[ a]], nv},
											(nv = Table[ VAR$[ {$varIndex++, Null}], {arity - la}];
											Lambda$TM[ RNG$ @@ (SIMPRNG$ /@ nv), True, Join[ e, # @@ nv]]) /; la < arity
										]
						  	]&,
						  	ho
						],
			
			_,
			ho = Select[ v,
					Function[ var,
						tmp = Cases[ newExpr, var[ args___] :> Hold[ args], {0, Infinity}, Heads -> True];
						If[ MemberQ[ tmp, Hold[ ___, _?isFreeSeqTerm[ g], ___]],
							False,
							tmp = DeleteDuplicates[ Length /@ tmp];
							If[ Length[ tmp] === 1 && First[ tmp] > 0 && FreeQ[ newExpr //. var[ args___] :> Hold[ args], var],
								ar[ var] = First[ tmp];
								True,
								False
							]
						]
					]
				]
		];
				
		(* For the instantiations, we use the meta-level $TMHOR$Lambda instead of the object-level Lambda$TM.
		$TMHOR$Lambda has only two arguments (no condition). *)
		rules = Map[ With[ {arity = ar[ #]},
						If[ MatchQ[ #, VAR$[ {_, _SEQ0$|_SEQ1$}]],
							SEQ$[ #] :>
								Module[ {nv},
									nv = Table[ VAR$[ {$varIndex++, Null}], {arity}];
									$TMHOR$Lambda[ RNG$ @@ (SIMPRNG$ /@ nv), SEQ$[ # @@ nv]]
								],
										
							# :>
								Module[ {nv},
									nv = Table[ VAR$[ {$varIndex++, Null}], {arity}];
									$TMHOR$Lambda[ RNG$ @@ (SIMPRNG$ /@ nv), # @@ nv]
								]
						]
					  ]&,
					  ho
				];
		
		Join[ newExpr, {Replace[ is, rules, {1}], ho}]
	]
	
isFreeSeqTerm[ g_List][ expr_] :=
	With[ {h = getVarHead[ expr]},
		MatchQ[ h, VAR$[ {_, _SEQ0$|_SEQ1$}]] && MemberQ[ g, h]
	]


(* ::Section::Closed:: *)
(* Methods for turning Formulas into {LHS, RHS, Condition} *)


(* 'decomposeFormula[ form, decompEquiv]' returns a tree corresponding to the decomposition of 'form'
into individual subparts (possibly) giving rise to rewrite-rules. Each leaf, is of the
form '{expr, conds, vars, prf, type}', where 'expr' is a subpart of 'form', 'conds' and
'vars' are the conditions and variables, respectively, in whose scope 'expr' appears, 'prf' is the
proof for obtaining 'expr' from 'from', and 'type' is a pattern indicating which types of rules shall
be generated from 'expr'. Nodes, i.e. heads of subparts, are of a similar form, except that they are
lacking the 'type'-entry.
The head of the whole output is guaranteed to have no free variables, i.e. it is of the form
'{expr, conds, {}, prf, (type)}', but it might have a non-empty list of conditions; this in particular
implies that the overall head does not necessarily have to correspond 1-1 to the original formula. *)
decomposeFormula[ form_, decompEquiv_] :=
	With[ {tmp = decomposeFormula[ form, {}, {}, Hold[], decompEquiv, True], h = {_, _, {}, _Hold, ___}},
		If[ MatchQ[ tmp, {h}|{h[ ___]}],
			First[ tmp],
			
			{form, {}, {}, Hold[]} @@ tmp
		]
	]
decomposeFormula[ form:Forall$TM[ r_, c_, f_], vars_, cond_, prf_, decompEquiv_, genFwd_] :=
	With[ {rc = makeUniqueVarNames[ rngToCondition[ r]], rv = rngVariables[ r]},
		(* We need to call 'makeUniqueVarNames' again, because 'rngToCondition' might introduce new binders. *)
		
		Module[ {newCond = Join[ cond, rc], newPrf = Join[ Append[ prf, {"Inst", Replace[ rv, VAR$[ {i_, n_}] :> FIX$[ n, i], {1}]}], makeMPTac[ rc]]},
			Switch[ c,
				_And$TM,
				newCond = Join[ newCond, List @@ c];
				newPrf = Join[ newPrf, makeMPTac[ c]],
				Except[ True],
				AppendTo[ newCond, c];
				AppendTo[ newPrf, "MP"]
			];
			{{form, DeleteDuplicates[ cond], vars, prf} @@ decomposeFormula[ f, Join[ vars, rv], newCond, newPrf, decompEquiv, genFwd]}
		]
	]
decomposeFormula[ Implies$TM[ l_, r_], vars_, cond_, prf_, decompEquiv_, genFwd_] :=
	decomposeFormula @@ Join[
		{r, vars},
		Switch[ l,
			_And$TM,
			{Join[ cond, List @@ l], Join[ prf, makeMPTac[ l]]},
			True,
			{cond, Append[ prf, "MP"]},
			_,
			{Append[ cond, l], Append[ prf, "MP"]}
		],
		{decompEquiv, genFwd}
	]
decomposeFormula[ And$TM[ conj_], vars_, cond_, prf_, decompEquiv_, genFwd_] :=
	decomposeFormula[ conj, vars, cond, Append[ prf, "ConjE"], decompEquiv, genFwd]
decomposeFormula[ form:And$TM[ _, __], vars_, cond_, prf_, decompEquiv_, genFwd_] :=
	With[ {sub = Join @@ MapIndexed[
					With[ {i = First[ #2]}, decomposeFormula[ #1, vars, cond, Join[ prf, Hold[ "ConjE", {"MvFrnt", i}]], decompEquiv, False]]&,
					form	
				]
			},
		If[ TrueQ[ genFwd],
			Append[
				sub,
				{form, DeleteDuplicates[ cond], vars, prf, "Forward"}
			],
			sub
		]
	]
decomposeFormula[ form:(Iff$TM[ l_, mid___, r_]), vars_, cond_, prf_, decompEquiv_, genFwd_] :=
	Module[ {sub, suffix},
		If[ TrueQ[ decompEquiv],
			suffix = If[ {mid} === {}, Hold[ "EquivE"], Hold[ "EquRST", "EquivE"]];
			sub = Join[
					decomposeFormula[ Implies$TM[ l, r], vars, cond, Join[ prf, suffix], decompEquiv, genFwd],
					decomposeFormula[ Implies$TM[ r, l], vars, cond, Join[ prf, suffix, Hold[ {"MvFrnt", 2}]], decompEquiv, genFwd]
				],
			
			sub = {}
		];
		Prepend[ sub, {form, DeleteDuplicates[ cond], vars, prf, If[ TrueQ[ genFwd], Blank[], Except[ "Forward"]]}]
	]
decomposeFormula[ form_, vars_, cond_, prf_, _, True] :=
	{{form, DeleteDuplicates[ cond], vars, prf, Blank[]}}
decomposeFormula[ form_, vars_, cond_, prf_, _, _] :=
	{{form, DeleteDuplicates[ cond], vars, prf, Except[ "Forward"]}}
	
makeMPTac[ c_And$TM] :=
	With[ {n = Max[ 1, Length[ c]]},
		Join[ Hold[ "ImpE", "ConjI"],
			Switch[ n,
				1,
				Hold[ "Triv"],
				_,
				Hold[ {"RepeatNTac", "Triv", n}]
			]
		]
	]
makeMPTac[ {}] :=
	Hold[]
makeMPTac[ c:{_, __}] :=
	With[ {n = Length[ c]},
		Hold[ "ImpE", "ConjI", {"RepeatNTac", "Triv", n}]
	]
makeMPTac[ _] :=
	Hold[ "MP"]
	
composeFormula[ $Failed, ___] :=
	{$Failed, Hold[]}
composeFormula[ inner_, {c_}, vars_, prf1_Hold, prf2_Hold] :=
	composeFormula[ Implies$TM[ c, inner], {}, vars, Prepend[ prf1, "ImpI"], prf2]
composeFormula[ inner_, conds:{_, __}, vars_, prf1_Hold, prf2_Hold] :=
	composeFormula[ inner, {makeConjunction[ conds, And$TM]}, vars, Prepend[ prf1, "ConjE"], prf2]
composeFormula[ inner_, {}, vars_List, prf1_Hold, prf2_Hold] :=
	Module[ {out = inner, prf = prf1},
		If[ prf =!= Hold[], AppendTo[ prf, {"MvFrnt", -1}]];
		If[ vars =!= {},
			out = Forall$TM[ RNG$ @@ (SIMPRNG$ /@ vars), True, out];
			PrependTo[ prf, {"AllI", Replace[ vars, VAR$[ {i_, n_}] :> FIX$[ n, i], {1}]}]
		];
		{postprocessExpression[ out], DeleteCases[ Join[ prf, prf2], {"MvFrnt", 1}]}
	]

makeHORules[ {form:(EqualDef$TM|Equal$TM)[ l_, r:(_Such$TM|_SuchUnique$TM)], c_List, glob_List, ho_List, prf_}, "ImplicitDefinition", _, _] :=
	With[ {tmp = makeHORule[ l, r, c, ho, glob, form, Append[ prf, "Triv"]]},
		{
			Switch[ tmp,
				_List,
				Append[ tmp, Mode -> "implicit"],
				_,
				$Failed
			]
		}
	]
makeHORules[ {form:IffDef$TM[ l_, r_], c_List, glob_List, ho_List, prf_}, type:("LazyDefinition"|"ExplicitDefinition"), lazy_, _] :=
	With[ {ls = MemberQ[ lazy, Head[ l]]},
		Which[
			type === "LazyDefinition" && ls,
			With[ {tmp = makeHORule[ l, r, c, ho, glob, form, Append[ prf, "Triv"]]},
				{
					Switch[ tmp,
						_List,
						Append[ tmp, Mode -> "lazy"],
						_,
						$Failed
					]
				}
			],
			
			type === "ExplicitDefinition" && !ls,
			{makeHORule[ l, r, c, ho, glob, form, Append[ prf, "Triv"]]},
			
			True,
			{}
		]
	]
makeHORules[ {form:EqualDef$TM[ l_, r:Except[_Such$TM|_SuchUnique$TM]], c_List, glob_List, ho_List, prf_}, "ExplicitDefinition", _, _] :=
	{makeHORule[ l, r, c, ho, glob, form, Append[ prf, "Triv"]]}
makeHORules[ {(h:(Equal$TM|Iff$TM))[ l:Except[ _Such$TM|_SuchUnique$TM], mid___, r:Except[ _Such$TM|_SuchUnique$TM]], c_List, glob_List, ho_List, prf_}, "Substitution", _, r2l_] :=
	Module[ {newPrf = prf, out},
		If[ {mid} =!= {},
			AppendTo[ newPrf, "EquRST"]
		];
		out = {makeHORule[ l, r, c, ho, glob, h[ l, r], Append[ newPrf, "Triv"]]};
		If[ NumberQ[ l] || TrueQ[ r2l],
			AppendTo[ out, makeHORule[ r, l, c, ho, glob, h[ r, l], Join[ newPrf, Hold[ "EquSym", "Triv"]]]]
		];
		out
	]
makeHORules[ {form_, c_List, glob_List, ho_List, prf_}, "Forward", _, _] :=
	Module[ {aPrf, negAssm, negConcl, newPrf1, newPrf2, repl = Replace[ glob, v:(VAR$[ {i_, n_}]) :> (v -> FIX$[ n, i]), {1}]},
		Switch[ c,
			{},
			negAssm = False;
			aPrf = Hold[],
			
			{_},
			With[ {assm = First[ c]},
				{negAssm, aPrf} = simplifiedNotProof[ Not$TM[ assm], Forward];
				PrependTo[ aPrf, {"Cases", assm}]
			],
			
			_,
			With[ {assm = makeConjunction[ c, And$TM]},
				{negAssm, aPrf} = Transpose[ simplifiedNotProof[ Not$TM[ #], Forward]& /@ c];
				negAssm = Or$TM @@ negAssm;
				aPrf = Join[
						Hold[ {"Cases", assm}, {"NotConjD", 1}, "DisjE"],
						Sequence @@ MapIndexed[ With[ {i = First[ #2]}, Prepend[ #1, {"DisjI", i}]]&, aPrf],
						Hold[ "ConjE"]
					   ]
			]
		];
		Join[
			MapIndexed[ makeHORule[ #1, form, Drop[ c, #2], ho, glob, Implies$TM[ #1, form], Hold[ "ImpI"], Append[ prf, "Triv"]]&, c],
			Switch[ form,
				And$TM[ __],
				MapIndexed[
					(
						{negConcl, newPrf1, newPrf2} = makeContrapositionProof[ #1, First[ #2], aPrf];
						makeHORule[ negConcl, negAssm, {}, ho, glob, Implies$TM[ negConcl, negAssm], newPrf1 /. repl, Join[ prf, newPrf2]]
					)&,
					List @@ form
				],
				_,
				{negConcl, newPrf1, newPrf2} = makeContrapositionProof[ form, 0, aPrf];
				{makeHORule[ negConcl, negAssm, {}, ho, glob, Implies$TM[ negConcl, negAssm], newPrf1 /. repl, Join[ prf, newPrf2]]}
			]
		]
	]
makeHORules[ {form_, c_List, glob_List, ho_List, prf_}, "Backward", _, _] :=
	With[ {fl = freeVariables[ form]},
		Module[ {conj = makeConjunction[ c, And$TM], ex = Complement[ freeVariables[ c], fl], prf1 = If[ Length[ c] > 1, Hold[ {"ConjE"}], Hold[]]},
			If[ ex =!= {},
				conj = Exists$TM[ RNG$ @@ Map[ SIMPRNG$, ex], True, conj];
				PrependTo[ prf1, {"ExE", Replace[ ex, VAR$[ {i_, n_}] :> FIX$[ n, i], {1}]}]
			];
			{makeHORule[ form, conj, {}, Select[ ho, (!MemberQ[ ex, #])&], Select[ glob, (!MemberQ[ ex, #])&], Implies$TM[ conj, form], Prepend[ prf1, {"ImpI"}], Append[ prf, "Triv"], fl]}
		]
	]
makeHORules[ ___] := {}

makeContrapositionProof[ concl_, index_, aPrf_] :=
	Module[ {negConcl, cPrf},
		{negConcl, cPrf} = simplifiedNotProof[ Not$TM[ concl], Backward];
		If[ cPrf =!= Hold[], cPrf = Join[ Hold[ {"SbGl", Not$TM[ concl]}], cPrf, Hold[ {"Remove", 2}]]];
		{
			negConcl,
			Join[ Hold[ "ImpI"], cPrf, aPrf],
			If[ index === 0, Hold[ {"Contr", 1}], Hold[ "ConjE", {"Contr", index}]]
		}
	]

(* 'makeHORule' does NOT call 'generateHORule', but only returns the arguments that have to be passed to it
(except optional arguments), as well as some additional information.
The result of 'makeHORule', thus, is either a pair of the form '{{lhs, rhs, conds, ho, vars}, {form, prf}}' or $Failed. *)
makeHORule[ _?NumberQ, _, _List, _List, _List, ___] :=
	$Failed
makeHORule[ a_, a_, _List, _List, _List, ___] :=
	$Failed (* We do not replace expressions by identical ones *)
makeHORule[ l_, r_, c_, ho_, glob_, form_, prf_] :=
	makeHORule[ l, r, c, ho, glob, form, Hold[], prf]
makeHORule[ l_, r_, c_, ho_, glob_, form_, prf1_, prf2_] :=
	With[ {fl = freeVariables[ l]},
		If[ Complement[ freeVariables[ r], fl] =!= {},
			$Failed,
			makeHORule[ l, r, c, ho, glob, form, prf1, prf2, fl]
		]
	]
makeHORule[ l_, r_, c_List, ho_List, glob_List, form_, prf1_, prf2_, fl_] :=
	Module[ {addVars, remVars, newC, newPrf1, newPrf2, newGlob},
		If[ Complement[ fl, glob] === {},
			addVars = Complement[ freeVariables[ c], fl];
			Switch[ addVars,
				{},
				newC = c;
				newPrf1 = Hold[],
				_,
				newC = {Exists$TM[ RNG$@@Map[ SIMPRNG$, addVars], True, makeConjunction[ c, And$TM]]};
				newPrf1 = Hold @@ {{"ExE", Replace[ addVars, VAR$[ {i_, n_}] :> FIX$[ n, i], {1}]}};
				If[ Length[ c] > 1, AppendTo[ newPrf1, {"ConjE"}]]
			];
			remVars = Complement[ glob, fl, addVars];
			newPrf2 =
				Replace[
					prf2,
					{"Inst", i__} :>
					With[ {o = Prepend[ Replace[ {i}, FIX$[ n_, j_] /; MemberQ[ remVars, VAR$[ {j, n}]] -> $TMHOR$Unknown, {1}], "Inst"]},
						o /; True
					],
					{1}
				];
			remVars = Join[ remVars, addVars];
			newGlob = Select[ glob, (!MemberQ[ remVars, #])&];
			{{l, r, newC, Select[ ho, (!MemberQ[ remVars, #])&], newGlob}, composeFormula[ form, newC, newGlob, Join[ newPrf1, prf1], newPrf2]},
		
			$Failed
		]
	]
makeHORule[ ___] := $Failed


(* ::Section:: *)
(* Methods for turning {LHS, RHS, Condition} into Rewrite Rules *)


(* TODO: Ensure that 'l' and 'r' are individuals. *)
Options[ generateHORule] = {SignalFailure -> False, Reference -> Null, ReferenceID -> -1, Mode -> "default"};
generateHORule[ l_, r_, c_List, v_List, glob_List, opts___?OptionQ] :=
	Module[ {rhs, lhs, rbp, ho, ms, allVars, randRanges, sf, ref, mode, strict, checkF, dCheckF,
				type, sow, out, nal, patt, mvars, cond, body, k, is, ce, cc, ci, rid, tokenRepl},
		{randRanges, sf, ref, mode, strict, checkF, dCheckF, type, out, sow, nal, rid} =
			{RandomRanges, SignalFailure, Reference, Mode, Strict, ImplicitCheckingFunction, LazyCheckingFunction, Type, RuleOutput, RuleSow, NormalizeArgumentLists, ReferenceID}
				/. {opts} /. Join[ Options[ generateHORule], Options[ formulaToHORules]];
				
		k = Switch[ ref,
				_FML$,
				key[ ref],
				_,
				Null
			];
			
		tokenRepl = {
						"NewExpression" :> First[ TMP$],
						"Condition" :> TMP$[[2]],
						"Instantiation" :> Last[ TMP$],
						"Formula" -> ref,
						"Key" -> k,
						"ReferenceID" -> rid
					};
	
		If[ mode =!= "lazy", dCheckF = Replace[ LazyCheckingFunction, Options[ formulaToHORules]]];
		Switch[ type,
			"Computation",
			sow = {};
			{ce, cc, ci} = {True, True, False},
			
			_,
			sow = Cases[ sow, HoldPattern[ _String -> _]];
			{ce, cc, ci} = Map[ Function[ t, MemberQ[ out, t, {0, Infinity}] || MemberQ[ sow, _ -> _?(MemberQ[ #, t, {0, Infinity}]&)]], {"NewExpression", "Condition", "Instantiation"}]
		];
		
		Catch[
			(* 'rhs' is a list containing both the real right-hand-side of the rewrite rule and all conditions. *)
			Which[
				mode === "implicit" && type === "Proof",
				Switch[ r,
					Such$TM[ _RNG$, _, _],
					rhs = $TMHOR$ImplDef@@r,
					SuchUnique$TM[ _RNG$, _, _],
					rbp = rngVariables[ First[ r]];
					ms = Last[ r] /. Map[ (# -> VAR$[ {$varIndex++, #[[1, -1]]}])&,
										DeleteDuplicates[ Cases[ Last[ r], VAR$[ {_Integer, _}]?((!MemberQ[ glob, #] && !MemberQ[ rbp, #])&), {0, Infinity}, Heads -> True]]
									];
					allVars = Map[ (# -> VAR$[ {$varIndex++, Null}])&, rbp];
					rhs = ReplacePart[
								$TMHOR$ImplDef@@r,
								3 -> And$TM[ Last[ r], Forall$TM[ Sequence@@(Most[ r] /. allVars),
												Implies$TM[ ms /. allVars, makeConjunction[ Map[ Equal$TM[ Last[ #], First[ #]]&, allVars], And$TM]]
											]
										]
						],
					_,
					Throw[ $Failed, ReturnValue]
				];
				rhs = If[ checkF === None,
						{rhs, simplifiedAnd[ makeConjunction[ c, And$TM]]},
						{rhs, c}
					],
					
				type === "Computation",
				(* computations require a LIST of conditions *)
				rhs = {r, c},
					
				True,
				rhs = {r, simplifiedAnd[ makeConjunction[ c, And$TM]]}
			];
			If[ ci,
				is = Replace[ glob, var:VAR$[ {_, _SEQ0$|_SEQ1$}] :> SEQ$[ var], {1}],	(* We need to wrap 'SEQ$' around every sequence variable. *)
				is = {}
			];
			
			(* Transform 'l' and 'rhs' such that every higher-order variable is always applied to the same number of arguments. *)
			{lhs, rhs, is, ho} = normalizeArities[ l, rhs, is, v, glob, nal];
			
			patt = mmaTransHORule[ lhs, Append[ rhs, is], ho, glob, randRanges, sf, strict, dCheckF, ce, cc];
			Switch[ patt,
				{_, _, _, _},
				{patt, mvars, cond, body} = patt,
				_,
				Throw[ $Failed, ReturnValue]
			];
			
			Switch[ type,
				"Computation",
				(*
					We construct a string matching the following pattern, depending on whether there is a semantic condition or 'cond' is 'True' or not:
					
					semantic condition or 'cond' === 'True':
						"patt /; kbSelectCompute[ ref] := Module[ mvars, trackResult[ First[ TMP$], ref] /; (body && trackCondition[ TMP$[[2]], lhs])]"
					
					no semantic condition and 'cond' =!= 'True':
						"patt /; kbSelectCompute[ ref] && cond := Module[ mvars, body; trackCondition[ {}, lhs]; trackResult[ First[ TMP$], ref]]"
						
					Note that for computations 'ref' must not be 'Null'.
				*)
				(* TODO: Implement separate versions of runtime functions for "Computation`" context.
					These versions should prevent unwanted evaluation, but also take into account that constituents of the new expression have
					'Hold' attributes themselves.
					The problem is that most/all of these functions need Hold-attribtes, because they might be called on expressions shielded by
					quantifiers; however, local program variables are not evaluated either then. *)
				Switch[ ref,
					_FML$,
					If[ Length[ patt] === 2 && Head[ patt] === Condition,
						patt = ReplacePart[ patt, 2 -> With[ {ky = k, cnd = Last[ patt]}, HoldComplete[ And[ Theorema`Common`kbSelectCompute[ ky], cnd]]]],
						patt = With[ {ky = k}, Condition[ patt, Theorema`Common`kbSelectCompute[ ky]]]
					];
					mvars = ReplacePart[ HoldComplete@@{mvars}, {1, 0} -> List];
					body = With[ {reference = ref, lOrig = postprocessExpression[ lhs], prg = body},
						If[ c === {} && cond === False,
							Join[ prg, HoldComplete[ Theorema`Common`trackCondition[ {}, lOrig], Theorema`Common`trackResult[ First[ TMP$], reference]]],
							Condition[ HoldComplete[ Theorema`Common`trackResult[ First[ TMP$], reference]], And[ prg, Theorema`Common`trackCondition[ Part[ TMP$, 2], lOrig]]]
						]
					];
					
					body = Replace[ With[ {mv = mvars, prg = body},
										HoldComplete[ Module[ mv, prg]]
									],
									{HoldComplete[ a_] :> a, HoldComplete[ a___] :> CompoundExpression[ a], Hold[ a___] :> a},
									{1, Infinity},
									Heads -> True
							];
					body = With[ {expr = Unevaluated@@body}, ToString[ expr, InputForm]];
					patt = Replace[ With[ {p = patt}, HoldComplete[ p]],
									{HoldComplete[ a_] :> a, HoldComplete[ a___] :> CompoundExpression[ a], Hold[ a___] :> a},
									{1, Infinity},
									Heads -> True
							];
					patt = With[ {expr = Unevaluated@@patt}, ToString[ expr, InputForm]];
					
					Throw[ StringReplace[ patt <> ":=" <> body,
											{"Theorema`Language`" -> "Theorema`Computation`Language`", "Theorema`Knowledge`" -> "Theorema`Computation`Knowledge`"}
							],
							ReturnValue
					],
					
					_,
					Throw[ $Failed, ReturnValue]
				],
				
				_,
				(*else: type = Proof*)
				(*
					We construct a Mathematica transformation rule matching the following pattern, depending on whether 'cond' is 'True' or not:
					
					cond === True:
						patt :> Module[ mvars, (Sow[ val1, key1]; Sow[ val2, key2]; ...; out) /; body]
					
					cond =!= True:
						patt :> Module[ mvars, body; Sow[ val1, key1]; Sow[ val2, key2]; ...; out]
						
					where the val1, val2 etc., together with their corresponding key1, key2 etc., are as specified in 'RuleSow',
					and 'out' is as specified in 'RuleOutput'.
					In the val1, val2, out etc., all occurrences of
					- "NewExpression" are replaced by 'First[ TMP$]',
					- "Condition" by 'TMP$[[2]]',
					- "Instantiation" by 'Last[ TMP$]',
					- "Formula" by 'ref', and
					- "Extraction" by ???.
					
					If the RHS is an implicit definition and 'checkF' is not 'None', then a call to 'checkF' is added as well.
				*)
				sow = Replace[ sow, (key_ -> val_) :> (key -> (HoldComplete[ val] /. tokenRepl)), {1}];
				sow = Flatten[ HoldComplete@@Replace[ sow, (key_ -> HoldComplete[ val_]) :> HoldComplete[ Sow[ val, key]], {1}], 1, HoldComplete];
				
				(*
				ref = If[ ref === Null,
							HoldComplete[],
							With[ {reference = ref}, HoldComplete[ Sow[ reference, "ref"]]]
						];
				If[ ci,
					ref = Join[ ref, HoldComplete[ Sow[ Last[ TMP$], "inst"]]];
				];
				ref = Join[ ref, HoldComplete[ Sow[ Part[ TMP$, 2], "cond"]]];
				*)
				
				With[ {out0 = out, cf = checkF},
					If[ mode === "implicit" && cf =!= None,
						AppendTo[ mvars, HoldComplete[ out$]];
						out = HoldComplete[ If[ (out$ = cf[ Part[ TMP$, 2], KB$]) =!= False, Append[ out0, out$], {}]],
						
						out = HoldComplete[ out0]
					]
				];
				out = Join[ sow, out /. tokenRepl];
				
				mvars = ReplacePart[ HoldComplete@@{mvars}, {1, 0} -> List];
				
				body = If[ cond,
							Condition@@{out, body},
							Join[ body, out]
						];
				
				Throw[ {
						k,
						With[ {mv = mvars, prg = body},
							Switch[ mode,
								"implicit",
								Switch[ checkF,
									None,
									{patt, patt :> Module[ mv, prg]},
									_,
									{patt, {patt, Pattern[ KB$, Blank[ List]]} :> Module[ mv, prg]}
								],
								_,
								patt :> Module[ mv, prg]
							]
						] //. {HoldComplete[ a_] :> a, HoldComplete[ a___] :> CompoundExpression[ a], Hold[ a___] :> a},
						glob === {}
					},
					ReturnValue
				]
			],
			
			ReturnValue
		]
	]

mmaTransHORule[ l_, r:{_, _, _}, v_List, g_List, randRanges_, sf_, strict_, dCheckF_, constructExpr_, constructCond_] :=
	Module[ {vars = v, k, alpha, alphaList, lhs = l, rhs = r, glob = g, allVars, hoVars, inst2,
			locStruct, hoStruct, pc = $Failed, pattern, gen, linearHO, modVars, fresh, rbp, lbp, prog,
			init, ms, drop, suchRng},
	
		Catch[
			If[ TrueQ[ randRanges],
				{lhs, rhs, drop} = addRangeVariables[ lhs, rhs, vars, glob];
				vars = Join[ vars, drop];
				glob = Join[ glob, drop]
			];
			
			(* Maybe this is not the right place to do the following ... *)
			If[ constructExpr === False, rhs[[1]] = {}];
			If[ constructCond === False, rhs[[2]] = True];
			
			(* Insert separate VAR$[i1, i2, {arity, i3}, inst] for every occurrence of a higher-order variable VAR$[i1], on the LHS.
			'i3' is the alpha-list, with the meaning that two occurrences of higher-order variables get the same 'j'-th entry in
			their alpha-lists iff they are occurrences of the same variable and their 'j'-th arguments are alpha-equivalent.
			Note that two alpha-equivalent arguments of the same occurrence get the same entry in the alpha-list of the occurrence, too!
			Moreover, every negative entry in the alpha-list corresponds to a local variable as the argument. *)
			k = 0;
			allVars = {};
			alphaList = {};
			alpha = 1;
			lhs = lhs //. (v1:VAR$[ i1_])[ a__] /; MemberQ[ vars, v1] :>
				Module[ {nv, ai, list, tmp},
					list = Cases[ alphaList, (i1 -> lst_List) :> lst, {1}, 1];
					If[ list === {},
						AppendTo[ alphaList, i1 -> {}];
						list = {},
						list = First[ list];
					];
					ai = Map[(
						tmp = Cases[ list, (candidate_ -> index_) /; alphaEquivalent[ candidate, #, {}] :> index, {1}, 1];
						If[tmp === {},
							Switch[ #,
							_VAR$,
							If[ MemberQ[ glob, #],
								AppendTo[ list, # -> alpha];
								alpha++,
								AppendTo[ list, # -> -alpha];
								-(alpha++)
							],
							_,
							AppendTo[ list, # -> alpha];
							alpha++
							],
							First[ tmp]
						]
					)&, {a}];
					alphaList = Replace[ alphaList, (i1 -> _) -> (i1 -> list), {1}];
					nv = VAR$[ i1, k++, {Length[ Hold[ a]], ai}, False];
					AppendTo[ allVars, nv];
					nv[ a]
				];
			
			(* Insert separate VAR$[i1,i2,inst] for every occurrence of a global first-order variable VAR$[i1] on LHS.
			For each i1, start with i2=0, but if there is only one occurrence of the respective variable, replace it by VAR$[i1,-1,inst] *)
			Scan[(
				k = 0;
				lhs = lhs /. # :> With[ {nv = VAR$[ First[ #], k++, False]}, AppendTo[ allVars, nv]; nv];
				If[ k === 1,
					{lhs, allVars} = {lhs, allVars} /. VAR$[ First[ #], 0, False] -> VAR$[ First[ #], -1, False]
				]
				)&,
				Complement[ glob, vars]
			];
			
			While[ pc === $Failed,
				pattern = (lhs /. (var:VAR$[ _, _, _, _])[ ___] :> var) /. VAR$[ f__, False] :> VAR$[ f, True];
				gen = Cases[ pattern, _VAR$, {0, Infinity}, Heads -> True];
				{lhs, allVars} = {lhs, allVars} /. VAR$[ f__, False] /; MemberQ[ gen, VAR$[ f, True]] :> VAR$[ f, True];
				
				lbp = bindingPositions[ lhs];
				locStruct = getLocalStructure[ lhs, lbp];
				hoStruct = getHOStructure[ lhs];
				
				(* 'locStruct' is now modified such that every 'o -> {l, f, s}' is turned into 'o -> {l, f, s, c}', where 'c' is a list containing all local variables that 'o' must be checked to be free of.
				Here, it is also taken into account that 'o' might be in an argument of a higher-order variable that has already been checked to be free of some local variables. *)
				locStruct = Replace[ locStruct, (op_ -> {lp_, fp_, sp_}) :> (op -> {lp, fp, sp, Complement[ lp, fp, Union@@Map[ (Complement@@Take[ select[ locStruct, #], 2])&, select[ hoStruct, op]]]}), {1}];
				
				(* 'linearHO' contains the indices of all higher-order variables that occur only once on the LHS. *)
				linearHO = Select[ DeleteDuplicates[ Cases[ allVars, VAR$[ i_, _, _, _] :> i]], Count[ allVars, VAR$[ #, _, _, _]] === 1&];
				
				(* 'hoVars' contains all occurrences of higher-order variables that are not simple:
				A higher-order variable is simple iff it is linear and applied only on _distinct_ local variables. *)
				hoVars = Cases[ allVars, var:(_[ i_, _, {arity_, _}, _]) /; (!MemberQ[ linearHO, i] || FreeQ[ lhs, var[ args:(VAR$[ _]..)] /; Length[ DeleteDuplicates[ {args}]] === arity])];
				
				(* Generate all conditions on all shape- and alpha-equivalence conditions on global variables.
				For simple higher-order variables, no shape/alpha-equivalence conditions need to be generated. *)
				pc = makeShapeConditions[ lhs, hoVars, locStruct, sf || strict];
				
				If[ MatchQ[ pc, {$Failed, _List}],
					If[ TrueQ[ sf],
						Throw[ $Failed, ReturnValue],
						
						(* We remove all higher-order variables that cause problems.
							This might not be the best strategy, because removing only one could also solve problems with others. *)
						vars = DeleteCases[ vars, VAR$[ j_] /; MemberQ[ Last[ pc], j]];
						allVars = DeleteCases[ allVars, VAR$[ j_, _, _, _] /; MemberQ[ Last[ pc], j]];

						(* ATTENTION! Here we rely on the fact that instantiations have $TMHOR$Lambda rather than Lambda$TM.
						Simply replacing $TMHOR$Lambda by Lambda$TM below would lead to wrong behavior.
						Also note that $TMHOR$Lambda does not have a condition! *)
						rhs[[-1]] = Last[ rhs] //. Map[ If[ MatchQ[ #, {_, _SEQ0$|_SEQ1$}],
															$TMHOR$Lambda[ _, SEQ$[ VAR$[ #][ __]]] -> SEQ$[ VAR$[ #]],
															$TMHOR$Lambda[ _, VAR$[ #][ __]] -> VAR$[ #]
														]&,
														Last[ pc]
													];
										
						If[ TrueQ[ randRanges],
							{lhs, rhs} = {lhs, rhs} //. {RNDRNG$[ x_, VAR$[ j:{_, Null}, __][ ___]] /; MemberQ[ Last[ pc], j] :> SIMPRNG$[ x],
														And$TM[ pre___, _[ VAR$[ j:{_, Null}, __][ ___]], post___] /; MemberQ[ Last[ pc], j] :> And$TM[ pre, post]}
						];
						
						Scan[(
							k = 0;
							lhs = lhs /. VAR$[ #, _, _, gen0_] :> With[ {nv = VAR$[ #, k++, gen0]}, AppendTo[ allVars, nv]; nv];
							If[ k === 1,
								{lhs, allVars} = {lhs, allVars} /. VAR$[ #, 0, gen0_] :> VAR$[ #, -1, gen0]
							]
							)&, Last[ pc]
						];
						pc = $Failed
					]
				]
			];
			
			(* We now construct the additional arguments of function 'anyInstance'. *)
			inst2 = Replace[ Last[ rhs],
							{
								$TMHOR$Lambda[ _, VAR$[ x_][ ___]] :>
									With[ {p = Position[ pattern, VAR$[ x, ___], {0, Infinity}, 1]},
										Switch[ p,
											{_},
											With[ {args = List@@Extract[ lhs, First[ p]]},
												With[ {loc = Map[ Cases[ freeVariables[ #], y:VAR$[ _] :> (y -> ToExpression[ VarName[ y, Variable], InputForm, Hold])]&, args]},
													{
														constructPattern[ #, {}, "i", locStruct, SameFO -> False, LocalPattern -> True, MainPattern -> True]& /@ (args /. (y:VAR$[ _List, _, _, _])[ ___] :> y /. Flatten[ loc]),
														loc[[All, All, 2]],
														ToExpression[ VarName[ Extract[ pattern, First[ p]], Occurrence], InputForm, Hold]
													}
												]
											],
											_,
											{$TMHOR$Unknown}
										]
									],
								_ -> {$TMHOR$Unknown}
							},
							{1}
					];
			
			(* Generate the alpha-equivalence conditions for all higher-order global variables. *)
			pc = Join[ pc, makeAlphaEquivalentFree[ locStruct, hoVars]];
			
			(* Insert separate VAR$[i, $TMHOR$HO|$TMHOR$FO] for every occurrence of a global variable VAR$[i] on the RHS. *)
			rhs = rhs /. (v1:VAR$[ i1_]) /; MemberQ[ glob, v1] :> If[ MemberQ[ vars, v1], VAR$[ i1, $TMHOR$HO], VAR$[ i1, $TMHOR$FO]];
			
			(* 'ms' contains the variable indices of all first-order, and the occurrences of all higher-order variables 'makeSequence' has to be called on.
			At the moment, 'makeSequence' has to be called on _all_ variables / non-instantiated occurrences, no matter whether they appear on the RHS or not
			(because they could appear in old arguments of 'substitute'). *)
			ms = Join[ DeleteDuplicates[ Cases[ allVars, _[ i_, j_, inst_] /; !(j === -1 && inst) :> i]], Cases[ allVars, _[ _, _, _, False]]];
			
			(* Compute the positions of all variable bindings in 'rhs'. *)
			rbp = Sort[ bindingPositions[ rhs], (Length[ Last[ #1]] < Length[ Last[ #2]] || (Most[ Last[ #1]] === Most[ Last[ #2]] && Last[ Last[ #1]] < Last[ Last[ #2]]))&];
			
			(* Collect all runtime variables that have to be introduced in the 'Module' environment.
			This includes one 'POS$i$a' for each higher-order variable 'i' with arity >= 'a',
			one 'VarName[o, Occurrence]' for each occurrence 'o' of a global variable not included in 'gen',
			one 'VarName[v, Variable]' for each first-order global variable 'v',
			one 'VarName[o, Bound]' for each occurrence 'o' of a higher-order global variable,
			one 'VarName[v, Len]' for each higher-order global sequence variable 'v',
			and one 'LOC$i' for each local variable appearing anywhere in the rule but not in 'gen'. *)
			modVars =
			HoldComplete@@Join[ {Hold[ TMP$]},
				Map[ ToExpression[ VarName[ #], InputForm, Hold]&, Complement[ lbp[[All, 1]], gen]],
				Map[ ToExpression[ VarName[ First[ #]], InputForm, Hold]&, rbp], Cases[ allVars, var:(_[ __, False]) :> ToExpression[ VarName[ var, Occurrence], InputForm, Hold]],
				Map[ ToExpression[ VarName[ #, Variable], InputForm, Hold]&, DeleteDuplicates[ Cases[ allVars, VAR$[ i_, Except[ -1], _] :> VAR$[ i, -1, False]]]],
				Map[ ToExpression[ VarName[ #, Bound], InputForm, Hold]&, Cases[ allVars, VAR$[ _, _, _, _]]],
				Map[ ToExpression[ VarName[ #, Len], InputForm, Hold]&, DeleteDuplicates[ Cases[ allVars, VAR$[ i:{_, _SEQ0$|_SEQ1$}, _, _, _] :> i]]],
				Flatten[ Cases[ DeleteDuplicates[ Cases[ allVars, _[ i_, _, {arity_, _}, _] :> {i, arity}]], {i_, arity_} :> Table[ ToExpression[ VarName[ i, ArgPositions, a], InputForm, Hold], {a, arity}]]]
			];
			
			(* Construct the new RHS, where applications of higher-order variables to arguments are replaced by substitutions.
			In addition, generate an initial list 'fresh'. *)
			{fresh, rhs} = generateFreshVars[ constructExpression[ rhs, lhs, hoVars]];
			
			(* Add exceptions to the elements of 'fresh': For each new local variable, collect all local variables in whose scope the global variables in 'lhs' appear,
			and add them as exceptions: We already know that the global variables must be free of those local variables, so we do not have to check freeness again in 'GetVar'. *)
			fresh = Replace[ fresh, (x_ -> {vp_, rp_}) :> (x -> {Map[ {#, select[ locStruct, #][[1]]}&, vp], rp}), {1}];
			
			(* Sort the variables in 'fresh' such that every variable 'v1' bound inside the scope of the quantifier that binds 'v2' appears _after_ 'v2'. *)
			fresh = Sort[ fresh, Position[ rbp[[All, 1]], First[ #1], {1}, 1][[1, 1]] < Position[ rbp[[All, 1]], First[ #2], {1}, 1][[1, 1]]&];

			(* Make sure that no new local variable gets the name of another local variable in whose scope it occurs. *)
			fresh = MapThread[ Insert[ #1, Cases[ rbp, HoldPattern[ lv_ -> ps_] /; (First[ #1] =!= lv && isInScope[ Last[ #2], ps]) :> lv], {2, 2}]&, {fresh, rbp}];
			
			(* Simplify 'fresh' and already replace the local variables, whose names are known, in 'rhs'.
			Moreover, we can even simplify some 'Substitute', if the old arguments agree with the new ones (and there is only _one_ possible instance). *)
			fresh = simplifyFreshVars@@fresh;

			rhs = rhs //. RNDRNG$[ _, rng_] :> rng
						//. Cases[ fresh, HoldPattern[ _ -> _VAR$]]
						//. u_VAR$ :> ToExpression[ VarName[ u, Occurrence], InputForm, Hold]
						//. Hold[ $TMHOR$Substitute][ _, a_List, {{b_, _, a_}}, ___] :> b;
			fresh = DeleteCases[ fresh, HoldPattern[ _-> _VAR$]];
			
			If[ !FreeQ[ rhs, VAR$], Throw[ $Failed, ReturnValue]];
			
			If[ MatchQ[ rhs, {$TMHOR$ImplDef[ _, _, _], ___}],
				suchRng = rhs[[1, 1]];
				rhs = ReplacePart[ rhs, 1 -> {suchRng, simplifiedAnd[ makeConjunction[ Prepend[ List@@Rest[ First[ rhs]], HoldComplete[ $TMHOR$RangeToCondition][ suchRng]], And$TM]]}]
			];
			
			rhs[[3]] = MapThread[ Function[ {x, y}, Hold[ $TMHOR$MakeInstance] @@ Prepend[ y, x]], {Last[ rhs], inst2}];
			rhs = OptimizeExpression[ rhs, modVars];
			
			pattern = constructPattern[ pattern, {}, "", locStruct, dCheckF, MainPattern -> True, SameFO -> False];
			Switch[ pc,
				{},
				(* If 'pc' is empty, we can initialize the global variables directly with sequences;
				Otherwise, they would first be initialized with lists and then immediately turned into sequences, which is pointless. *)
				init = Map[ With[ {n = ToExpression[ VarName[ #, Variable], InputForm, Hold], occ = ToExpression[ VarName[ #, Occurrence], InputForm, Hold]},
								n -> Hold[ n = occ]
							]&, DeleteDuplicates[ Cases[ allVars, _[ _, Except[ -1], True]], First[ #1] === First[ #2]&]];
				(* If 'pc' is empty, we certainly do not need 'FreeQ' (all higher-order variables are simple). *)
				prog = HoldComplete@@Append[ Map[ makeFreshCode, fresh], With[ {newRhs = rhs}, Hold[ TMP$ = newRhs]]],
				
				_,
				init = Join[
					Map[ With[ {n = ToExpression[ VarName[ #, Variable], InputForm, Hold], occ = ToExpression[ VarName[ #, Occurrence], InputForm, Hold]},
							n -> Hold[ n = {occ}]
						]&, DeleteDuplicates[ Cases[ allVars, _[ _, Except[ -1], True]], First[ #1] === First[ #2]&]],
					Map[ With[ {n = ToExpression[ VarName[ #, Len], InputForm, Hold], occ = ToExpression[ VarName[ #, Occurrence], InputForm, Hold]},
							n -> Hold[ n = Length[ {occ}]]
						]&, DeleteDuplicates[ Cases[ allVars, _[ {_, _SEQ0$|_SEQ1$}, _, _, True]], First[ #1] === First[ #2]&]]];
				prog = ReplacePart[
						HoldComplete @@ {Hold @@ Append[ pc, HoldComplete@@Join[
																Map[ makeSeqCode, ms],
																Map[ makeFreshCode, fresh],
																With[ {newRhs = rhs},
																	{Hold[ TMP$ = newRhs]}
																],
																If[ FreeQ[ Most[ rhs], $TMHOR$Substitute],
																	{True},
																	{Hold[ FreeQ[ First[ TMP$], $TMHOR$Unknown] && FreeQ[ Part[ TMP$, 2], $TMHOR$Unknown]]}
																]
															]
														]
										},
							{1, 0} -> And
						]
			];
			
			modVars = Select[ modVars, !FreeQ[ prog, HoldPattern@@#]&] /. init //. Hold[ a_] :> a;
			
			If[ MemberQ[ hoStruct, _ -> {__}],
				If[ dCheckF === None, pattern = Pattern[ P$, pattern]];
				modVars = Join[ modVars, HoldComplete[ BND$ = boundVariables[ P$]]]
			];
			
			Throw[ {pattern, modVars, pc =!= {}, prog}, ReturnValue],
			
			ReturnValue
		]
	]


(* ::Subsection::Closed:: *)
(* Auxiliary Functions *)


SetAttributes[ OptimizeExpression, HoldRest];
OptimizeExpression[ ex_, vars_] :=
	Module[ {expr = ex, l, f, i = 0, init = {}},
		l = Sort[ Replace[ GatherBy[ Map[ {Extract[ expr, #], #}&,
											DeleteCases[ Position[ expr, _?(!FreeQ[ #, Hold[ $TMHOR$Substitute][ ___]]&)], {}]],
									First
							], 
						{{_} :> Sequence[], all:{{e_, _List}..} :> (e -> all[[All, 2]])},
						{1}
					],
					Depth[ First[ #1], Heads -> True] < Depth[ First[ #2 ], Heads -> True]&
			];
		While[ l =!= {},
			f = First[ l];
			With[ {v = ToExpression[ "TMP$" <> ToString[ i++], InputForm, Hold], val = First[ f], pos = Last[ f]},
				AppendTo[ vars, v];
				expr = ReplacePart[ expr, pos -> v];
				l = Rest[ l] /. val -> v;
				AppendTo[ init, Hold[ v = val]]
			]
		];
		If[ init === {},
			expr,
			ReplacePart[ HoldComplete@@{Append[ init, expr]}, {1, 0} -> CompoundExpression] //. Hold[ a_] :> a
		]
	]
	
select[ LOCSTRUCT$[ ___, v_ -> l_, ___], v:(VAR$[ _, _, __])] := l
select[ _LOCSTRUCT$, VAR$[ _, _, __]] := {{}, {}, {}, {}}
select[ ls_LOCSTRUCT$, VAR$[ i_, _]] :=
	Module[ {c = Cases[ ls, HoldPattern[ VAR$[ i, _, __] -> l_] :> l]},
		Switch[ c,
		{},
		{{}, {}, {}, {}},
		_,
		Map[ Apply[ Union, #]&, Transpose[ c]]
		]
	]
select[ HOSTRUCT$[ ___, v_ -> l_, ___], v_] := l
select[ _HOSTRUCT$, _ ] := {}

makeSeqCode[ i_List] :=
	With[ {n = ToExpression[ VarName[ VAR$[ i, $TMHOR$FO], Variable], InputForm, Hold]},
		Hold[ n = $TMHOR$MakeSequence[ n]]
	]
makeSeqCode[ v_VAR$] :=
	With[ {n = ToExpression[ VarName[ v, Occurrence], InputForm, Hold]},
		Hold[ n = $TMHOR$MakeSequence[ n]]
	]
	
makeFreshCode[ v_ -> {exprs_List, ex_List, pref_List}] :=
	With[ {n = ToExpression[ VarName[ v, Variable], InputForm, Hold],
			args = Map[ Function[ x,
						{{ToExpression[ VarName[ First[ x], Occurrence], InputForm, Hold]},
						 Map[ ToExpression[ VarName[ #, Variable], InputForm, Hold]&, Last[x]]}],
						exprs
					],
			exn = Map[ ToExpression[ VarName[ #, Variable], InputForm, Hold]&, ex],
			prefn = Map[ ToExpression[ VarName[ #, Variable], InputForm, Hold]&, pref],
			orig = v[[1, 2]]
		},
		Hold[ n = $TMHOR$GetVar[ args, exn, prefn, orig]]
	]
makeFreshCode[ v_ -> {exprs_List, ex_List}] :=
	With[ {n = ToExpression[ VarName[ v, Variable], InputForm, Hold],
			args = Map[ {ToExpression[ VarName[ #, Occurrence], InputForm, Hold]}&, exprs],
			exn = Map[ ToExpression[ VarName[ #, Variable], InputForm, Hold]&, ex],
			orig = v[[1, 2]]
		},
		Hold[ n = $TMHOR$FreshVar[ args, exn, orig]]
	]
	
simplifyFreshVars[ pre___, v_ -> {exprs_List, ex:{___, var_, ___}, pref:{___, var_, ___}}, post___] :=
	simplifyFreshVars[ pre, v -> {exprs, ex, Select[ pref, !MemberQ[ ex, #]&]}, post]
simplifyFreshVars[ pre___, v_ -> {exprs:{___, {_, {___, var_, ___}}, ___}, ex:{___, var_, ___}, pref_List}, post___] :=
	simplifyFreshVars[ pre, v -> {Map[ {First[ #], Complement[ Last[ #], ex]}&, exprs], ex, pref}, post]
simplifyFreshVars[pre___, v_ -> {exprs_List, ex_List, {prepref___, var_, __}}, post___] /; (!MemberQ[ ex, var] && And@@Map[ MemberQ[ #, var]&, Hold@@exprs[[All, 2]]]) :=
	simplifyFreshVars[ pre, v -> {exprs, ex, {prepref, var}}, post]
(* In the following two rules: if we fix the name of a new local variable v to, say, x, we must ensure that no other local variable w that needs to have a different name than v may be named x as well.
		Hence, x is added to the exception-lists of all these w. *)
simplifyFreshVars[ pre___, v_ -> {exprs_List, ex_List, {var_}}, post___] /; (!MemberQ[ ex, var] && And@@Map[ MemberQ[ #, var]&, Hold@@exprs[[All, 2]]]) :=
	simplifyFreshVars@@Insert[ Replace[ {pre, post}, (x_?(MemberQ[ ex, #]&) -> {a_, b_, c_}) :> (x -> {a, Append[ b, var], c}), {1}] /. v -> var, v -> var, Length[ {pre}] + 1]
simplifyFreshVars[ pre___, v_ -> {{}, ex_List, {var_, ___}}, post___] /; !MemberQ[ ex, var] :=
	simplifyFreshVars@@Insert[ Replace[ {pre, post}, (x_?(MemberQ[ ex, #]&) -> {a_, b_, c_}) :> (x -> {a, Append[ b, var], c}), {1}] /. v -> var, v -> var, Length[ {pre}] + 1]
simplifyFreshVars[ pre___, rule:(v1_ -> {exprs1_List, ex1_List, {a___, b___}}), mid___, v2_ -> {exprs2_List, ex2_List, {a___}}, post___] /;
		(sequenceTypeEqual[ v1, v2] && Complement[ ex2, ex1] === {} && !MemberQ[ ex1, v2]) :=
	Module[ {tmp},
		(
			simplifyFreshVars@@Insert[ Insert[ {pre, mid, post} /. v2 -> v1, rule, Length[ {pre}] + 1], v2 -> v1, Length[ {pre, mid}] + 2]
		) /;
		Catch[
			Scan[(
				tmp = Cases[ exprs1, {First[ #], l_List} :> l];
				Switch[ tmp,
					{_},
					If[ Complement[ Last[ #], First[ tmp]] =!= {}, Throw[ False]],
					_,
					Throw[ False]
				]
				)&,
				exprs2
			];
			True
		]
	]
simplifyFreshVars[ pre___, v1_ -> {exprs1_List, ex1_List, {a___}}, mid___, rule:(v2_ -> {exprs2_List, ex2_List, {a___, b___}}), post___] /;
		(sequenceTypeEqual[ v1, v2] && Complement[ ex1, ex2] === {} && !MemberQ[ ex2, v1]) :=
	Module[ {tmp},
		(
			simplifyFreshVars@@Insert[ Insert[ {pre, mid, post} /. v1 -> v2, rule, Length[ {pre}] + 1], v1 -> v2, Length[ {pre, mid}] + 2]
		) /;
		Catch[
			Scan[(
				tmp = Cases[ exprs2, {First[ #], l_List} :> l];
				Switch[ tmp,
					{_},
					If[ Complement[ Last[ #], First[ tmp]] =!= {}, Throw[ False]],
					_,
					Throw[ False]
				]
				)&,
				exprs1
			];
			True
		]
	]
simplifyFreshVars[ args___] :=
	Replace[ {args}, (v_ -> {exprs_List, ex_List, {}}) :> (v -> {exprs[[All, 1]], ex}), {1}]
	
generateFreshVars[ expr_] :=
	Module[ {qPos, tmp, lin = {}, nl = {}, remRng},
		tmp = expr //. Hold[ $TMHOR$Substitute][ a_, b_, c:{___, {_, _, _, _}, ___}, trust___] :> $dummy[ a, b, Map[ Delete[ #, 3]&, c], trust];
		qPos = Cases[ tmp, _[ r_RNG$, rest___] :> {rngVariables[ r], {r, rest}}, {0, Infinity}, Heads -> True];
		lin = Select[ qPos, FreeQ[ Last[ #], $dummy[ _, _, {_, __}, ___]]&];
		nl = Complement[ qPos, lin];
		
		lin =
		Flatten[ Replace[ lin,
			{v_List, {r_, rest___}} :> (
					remRng = Prepend[ r, $dummy];(* '$dummy' is immediately removed. *)
					Map[ (remRng = Rest[ remRng];
						With[ {l = { remRng, rest}},
							# -> {DeleteDuplicates[ Cases[ l, VAR$[ _, __], {0, Infinity}, Heads -> True]], varPriorityList[ Cases[ l, (# -> x_) :> x, {0, Infinity}, Heads -> True]]}
						]
					)&, v]
				), {1}
				], 1];
		nl =
		Flatten[ Replace[ nl, {v_List, {r_, rest___}} :> (
					remRng = Prepend[ r, $dummy];
					Map[ (remRng = Rest[ remRng];
						# -> {DeleteDuplicates[ Cases[ {remRng, rest}, VAR$[ _, __], {0, Infinity}, Heads -> True]], {}}
					)&, v]
				),{1}
				], 1];
		
		{Join[ lin, nl], expr //. Hold[ $TMHOR$Substitute][ a_, b_, c:{___, {_, _, _, _}, ___}, trust___] :> Hold[ $TMHOR$Substitute][ a, b, Map[ Most, c], trust]}
	]
	
varPriorityList[ l_List] := Sort[ Tally[ l], Last[ #1] > Last[ #2]&][[All, 1]]

constructExpression[ (v:VAR$[ i_, $TMHOR$HO])[ args___], know_, nonSimple_List] :=
	Module[ {candidates, x, y, tmp = {}, instOnly = False, le, repl},
		(* Find the best matches for 'v' in 'know'. *)
		candidates = Cases[ know, (var:VAR$[ i, _, _, _])[ a___] :> {var, ToExpression[ VarName[ var, Bound], InputForm, Hold], {a} /. (ho:VAR$[ _, _, _, _])[ ___] :> ho}, {0, Infinity}, Heads -> True];
		If[ candidates === {},
			$Failed,
			
			(*else*)
			(* We want to find _all_ candidates '{var,{a}}' such that '{args}' and '{a}' agree up to local equivalence _and_ 'var' is certainly instantiated.
			If such a candidate does not exist, we return _all_ candidates, with those where '{args}' and '{a}' agree put at the beginning. *)
			x = {args} /. (var:VAR$[ _, $TMHOR$HO])[ ___] :> var;
			
			(* 'tmp' becomes a list '{{o, bnd, args, {nl1 \[Rule] ol1, ...}}, ...}'. *)
			Scan[ Function[ c,
					y = Last[ c] /. {VAR$[ j_, _, _, _] :> VAR$[ j, $TMHOR$HO], VAR$[ j_, _, _] :> VAR$[ j, $TMHOR$FO]};
					{le, repl} = Transpose[ MapThread[ localEquivalent[ #1, #2, {}]&, {x, y}]];
					le = And@@le;
					repl = Replace[
								Replace[ GatherBy[ Flatten[ repl, 1], First], elem:{(var_ -> _)..} :> (var -> elem[[All, 2]]), {1}],
								{(var_ -> {(new_)...}) :> (var -> new), (var_ -> l_List) :> (le = False; var -> First[ Commonest[ l]])},
								{1}
							];
					y = Append[ c, repl];
					If[ le,
						If[ instOnly,
							If[ Last[ First[ c]], AppendTo[ tmp, y]],
							If[ Last[ First[ c]],
								tmp = {y};
								instOnly = True,
								PrependTo[ tmp, y];
							]
						],
						If[ !instOnly, AppendTo[ tmp, y]]
					];
				], candidates
			];
			
			Hold[ $TMHOR$Substitute][
										Table[ ToExpression[ VarName[ i, ArgPositions, a], InputForm, Hold], {a, Length[{args}]}],
										Map[ constructExpression[ #, know, nonSimple]&, {args}],
										tmp,
										!MemberQ[ nonSimple, VAR$[ i, _, _, _]]
									]
		]
	]
constructExpression[ expr_, know_, nonSimple_List] :=
	expr /. v:(VAR$[ _, $TMHOR$HO][ ___]) :> constructExpression[ v, know, nonSimple]
	
(* The difference between 'alphaEquivalent' and 'localEquivalent' is that the latter one ignores differences
in _all_ local variables of the same kind (individual, sequence), not only in those bound in the expressions it is called on. *)
localEquivalent[ v1_VAR$, v2_VAR$, rules_List] :=
	Module[ {first = Cases[ rules, (v1 -> x_) :> x, {1}, 1]},
		Switch[ first,
			{},
			If[ Length[ v1] === Length[ v2] === 1,
				If[ sequenceTypeEqual[ v1, v2],
					{True, Append[ rules, v1 -> v2]},
					{False, {}}
				],
				{v1 === v2, rules}
			],
			{v2},
			{True, rules},
			_,
			{False, {}}
		]
	]
localEquivalent[ {a___}, {b___}, rules_List] :=
	Module[ {newRules = rules, tmp},
		If[ Length[ Hold[ a]] === Length[ Hold[ b]],
			Catch[
				Scan[ Function[ x,
						{tmp, newRules} = localEquivalent[ First[ x], Last[ x], newRules];
						If[ tmp === False, Throw[ {False, {}}]]
					], Transpose[ {{a}, {b}}]
				];
				{True, newRules}
			],
			{False, {}}
		]
	]
localEquivalent[ q_[ r_RNG$, a__], q_[ s_RNG$, b__], rules_List] :=
	Module[ {newRules, out1, out2, l = Length[ r]},
		If[ Length[ Hold[ a]] === Length[ Hold[ b]] && l === Length[ s],
			newRules = MapThread[ If[ sequenceTypeEqual[ #1, #2], (#1 -> #2), $Failed]&, {rngVariables[ r], rngVariables[ s]}];
			If[ MemberQ[ newRules, $Failed],
				{False, {}},
				
				newRules = Join[ newRules, rules];
				{out1, out2} = localEquivalent[ {r, a}, {s, b}, newRules];
				{out1, Drop[out2, l]}
			],
			{False, {}}
		]
	]
localEquivalent[ g_[ a___], h_[ b___], rules_List] := localEquivalent[ {g, a}, {h, b}, rules]
localEquivalent[ a_, b_, rules_] := {a === b, rules}

(* 'getLocalStructure[ expr, bp ]' returns 'LOCSTRUCT$[o1 -> {b1, f1, s1}, ...]' where each occurrence 'oi' of every global (first- and higher-order) variable in 'expr' is mapped to:
	'bi' ... the list of all local variables in whose scope 'oi' lies,
	'fi' ... the list of all elements of 'bi' that appear free in the arguments of 'oi',
	'si' ... the list of all elements of 'fi' that appear outside the arguments of any other higher-order variable. We obviously have 'si' \[SubsetEqual] 'fi' \[SubsetEqual] 'bi'.
For occurrences of first-order variables, 'fi' and 'si' are of course empty. *)
getLocalStructure[ expr_, bPos_List] :=
	Module[ {fPos, hPos, out = LOCSTRUCT$[ ], v, args, l},
		fPos = Position[ expr, VAR$[ _, _, _], {0, Infinity}, Heads -> True];
		hPos = Position[ expr, VAR$[ _, _, _, _], {0, Infinity}, Heads -> True];
		Scan[ Function[ p,
				v = Switch[ p, {}, expr, _, Extract[ expr, p]];
				AppendTo[ out, v -> {Cases[ bPos, HoldPattern[ n_ -> ps_] /; isInScope[ p, ps] :> n], {}, {}}]
			], fPos
		];
		Scan[ Function[ p,
				Switch[ p,
					{___, 0},
					v = Switch[ p, {_}, expr, _, Extract[ expr, Most[ p]]];
					args = List@@v;
					v = Head[ v];
					l = Cases[ bPos, HoldPattern[ n_ -> ps_] /; isInScope[ p, ps] :> n];
					AppendTo[ out, v -> {l, Intersection[ freeVariables[ args], l], Intersection[ freeVariables[ args /. VAR$[ _, _, _, _][ ___] -> Null], l]}],
					_,
					v = Switch[ p, {}, expr, _, Extract[ expr, p]];
					AppendTo[ out, v -> {Cases[ bPos, HoldPattern[ n_ -> ps_] /; isInScope[ p, ps] :> n], {}, {}}]
				]
			], hPos
		];
		out
	]
	
(* 'getHOStructure[ expr ]' returns 'HOSTRUCT$[o1 -> {h1, h2, ...}, ...]'
where each occurrence 'oi' of every global (first- and higher-order) variable in 'expr' is mapped
to the list containing all occurrences 'hi' of higher-order variables in whose arguments 'oi' lies.
'h1' is the outermost higher-order occurrence, and so on.
If 'oi' does not appear in the arguments of higher-order variables, the list is of course empty. *)
getHOStructure[ expr_] :=
	Module[ {gPos, hPos},
		gPos = HOSTRUCT$@@Position[ expr, VAR$[ _, __], {0, Infinity}, Heads -> True];
		hPos = Sort[ Position[ expr, VAR$[ _, _, _, _][ __], {0, Infinity}, Heads -> True]];
		Map[ Function[ p,
			If[ p === {},
				expr -> {},
				Extract[ expr, p] -> Map[ Extract[ expr, Append[ #, 0]]&, Select[ hPos, isSubposition[ p, #, False]&]]
			]
		], gPos]
	]
	
isSubposition[ p_, p_, _] := True
isSubposition[ {a___, __}, {a___}, True] := True
isSubposition[ {a___, Except[ 0], ___}, {a___}, _] := True
isSubposition[ _, _, _] := False

(* 'isInScope[ p1, p2 ]' checkes whether the expression at position 'p1' appears in the scope of the variable
bound in the range at position 'p2'. Range in this context refers to SIMPRNG, SETRNG, ...
The difference between 'inScope' and 'isInScope' is that in the former the second argument refers to the
position of the bound variable itself, whereas in the latter the second argument refers to its range. *)
isInScope[ p_List, p_List] := True
isInScope[ {a___, b_, ___}, {a___}] := b =!= 0
isInScope[ {a___, b_, ___}, {a___, c_}] := b > c
isInScope[ {a___, b_, __}, {a___, c_, _}] := b > c
isInScope[ _, _] := False

(* 'constructPattern[ expr, noblanks, prefix, ls ]' turns 'expr' into a pattern 'patt'.
'patt' is essentially the same as 'expr', only that every variable in 'expr' that is not included in 'noblanks' is replaced by named blanks,
where the name is just the 'VarName' of the variable with prefix 'prefix'.
In addition, a pattern condition 'c' is created, containing all alpha-equivalence and freeness constraints on the variables appearing in 'expr'.
More precisely, occurrences of the same first-order variable must be alpha-equivalent,
occurrences of the same higher-order variable with alpha-equivalent formal arguments must be alpha-equivalent as well,
and all occurrences of the same higher-order sequence variable must be instantiated by sequences of the same length.
All variables are checked to be instantiated by appropriate terms, i.e. an individual variable may not be instantiated by sequence terms, etc.
All variables are checked to be free of the local variables according to 'ls'.
Note that 'constructPattern' can distinguish, by the optional parameter 'MainPattern',
whether the pattern corresponds to the main pattern of a rewrite rule
(i.e. where no module variables, like 'BND$', are available) or to a pattern appearing already in the body of a 'Module', where module variables are available. *)
Options[ constructPattern] = {SameFO -> True, LocalPattern -> False, MainPattern -> False};
constructPattern[ expr_, noBlanks_List, prefix___String, ls_LOCSTRUCT$, opts___?OptionQ] :=
	constructPattern[ expr, noBlanks, prefix, ls, None, opts]
constructPattern[ expr_, noBlanks_List, prefix___String, ls_LOCSTRUCT$, dCheckF_, opts___?OptionQ] :=
	Module[ {repl, vars = DeleteDuplicates[ Cases[ expr, _VAR$, {0, Infinity}, Heads -> True]], sfo, free,
			cl, new, mp, hovars, allHOOccs, ai, comp},
			
		{sfo, mp} = {SameFO, MainPattern} /. {opts} /. Options[ constructPattern];
		allHOOccs = Cases[ ls, ((var:_[ _, _, _, _]) -> _List) :> var];
		repl = Map[ varToPattern[ #, !MemberQ[ noBlanks, #], prefix, SameFO -> sfo]&, vars];
		new = expr //. r_RNG$ :> constructRangePattern[ r, noBlanks, prefix] //. repl /. $dummy -> RNG$;
		vars = Complement[ vars, noBlanks];
		hovars = GatherBy[ Cases[ vars, _[ _, _, _, _]], #[[3, 2]]&];
		vars = Cases[ vars, _[ _, _, _]];
		If[ mp,
			If[ sfo,
				cl = DeleteCases[ Map[ Function[ v,
							free = Map[ ToExpression[ VarName[ #], InputForm, HoldComplete]&, Last[ select[ ls, v]]];
							If[ free === {},
								Null,
								HoldComplete[ $TMHOR$FreeOf][ {ToExpression[ prefix <> VarName[ v, Variable], InputForm, HoldComplete]}, free]
							]
						], vars],
						Null
				],
				(*else*)
				
				cl = DeleteCases[ Map[ Function[ v,
							free = Map[ ToExpression[ VarName[ #], InputForm, HoldComplete]&, Last[ select[ ls, v]]];
							If[ free === {},
								Null,
								HoldComplete[ $TMHOR$FreeOf][ {ToExpression[ prefix <> VarName[ v, Occurrence], InputForm, HoldComplete]}, free]
							]
						], Cases[ vars, _[ _, -1, _]]],
						Null
				];
				vars = GatherBy[ DeleteCases[ vars, _[ _, -1, _]], First];
				cl = Join[ cl, DeleteCases[ Map[ Function[ v,
							free = Map[ ToExpression[ VarName[ #], InputForm, HoldComplete]&, Union@@Map[ Last[ select[ ls, #]]&, v]];
							If[ Length[ v] === 1,
								If[ free === {},
									Null,
									HoldComplete[ $TMHOR$FreeOf][ {ToExpression[ prefix <> VarName[ First[ v], Occurrence], InputForm, HoldComplete]}, free]
								],
								HoldComplete[ $TMHOR$AlphaEquivalentFree][ Map[ {ToExpression[ prefix <> VarName[ #, Occurrence], InputForm, HoldComplete]}&, v], free]
							]
						], vars],
						Null
					]
				]
			];
			cl = Join[ cl,
				DeleteCases[ Map[ Function[ v,
							free = Map[ ToExpression[ VarName[ #], InputForm, HoldComplete]&, Union@@Map[ Last[ select[ ls, #]]&, v]];
							If[ Length[ v] === 1,
									If[ free === {},
									Null,
									HoldComplete[ $TMHOR$FreeOf][ {ToExpression[ prefix <> VarName[ First[ v], Occurrence], InputForm, HoldComplete]}, free]
								],
								HoldComplete[ $TMHOR$AlphaEquivalentFree][ Map[ {ToExpression[ prefix <> VarName[ #, Occurrence], InputForm, HoldComplete]}&, v], free]
							]
						], hovars],
					Null
				]
			];
			hovars = GatherBy[ Cases[ hovars[[All, 1]], _[ {_, _SEQ0$|_SEQ1$}, _, _, _]], First];
			cl = Join[ DeleteCases[ Map[ Function[ v,
							If[ Length[ v] <= 1,
								Null,
								HoldComplete[ $TMHOR$SameLength][ Map[ {ToExpression[ prefix <> VarName[ #, Occurrence], InputForm, HoldComplete]}&, v]]
							]
						], hovars],
					Null
				],
				cl
			];
			If[ dCheckF =!= None,
				PrependTo[ cl, HoldComplete[ dCheckF[ P$]]];
				new = Pattern@@Append [Hold[ P$], new]
			],
			(*else*)
		
			If[ sfo,
				cl = Map[ Function[ v,
						HoldComplete[ $TMHOR$IsGround][ {ToExpression[ prefix <> VarName[ v, Variable], InputForm, HoldComplete]}, HoldComplete[ BND$]]
					],
					vars
				],
				(*else*)
				
				cl = Map[ Function[ v,
							HoldComplete[ $TMHOR$IsGround][ {ToExpression[ prefix <> VarName[ v, Occurrence], InputForm, HoldComplete]}, HoldComplete[ BND$]]
						],
						Cases[ vars, _[ _, -1, _]
					]
				];
				vars = GatherBy[ DeleteCases[ vars, _[ _, -1, _]], First];
				cl = Join[ cl, Map[ Function[ v,
							HoldComplete[ $TMHOR$AlphaEquivalentGround][
								Prepend[ Map[ {ToExpression[ prefix <> VarName[ #, Occurrence], InputForm, HoldComplete]}&, v],
										ToExpression[ VarName[ First[ v], Variable], InputForm, HoldComplete]
								], HoldComplete[ BND$]]
						], vars]
				]
			];
			cl = Join[ cl, DeleteCases[ Map[ Function[ v,
							ai = v[[1, 3, 2]];
							comp = Complement[ Cases[ allHOOccs, _[ _, _, {_, ai}, _]], v];
							free = Map[ ToExpression[ VarName[ #], InputForm, HoldComplete]&, Union @@ Map[ select[ ls, #][[2]]&, v]];
							(* 'free' contains those local variables that may occur free in the instance of 'VarName[ #, Occurrence]'. *)
							If[ comp === {},
								If[ Length[ v] === 1,
									HoldComplete[ $TMHOR$IsGround][ {ToExpression[ prefix <> VarName[ First[ v], Occurrence], InputForm, HoldComplete]}, {HoldComplete[ BND$], free}],
									HoldComplete[ $TMHOR$AlphaEquivalentGround][ Map[ {ToExpression[ prefix <> VarName[ #, Occurrence], InputForm, HoldComplete]}&, v], {HoldComplete[ BND$], free}]
								],
								ai = Cases[ comp, _[ _, _, _, True], {1}, 1];
								If[ ai === {},
									HoldComplete[ $TMHOR$AlphaEquivalentGround][
										Map[ ToExpression[ VarName[ #, Occurrence], InputForm, HoldComplete]&, comp],
										Map[ {ToExpression[ prefix <> VarName[ #, Occurrence], InputForm, HoldComplete]}&, v],
										{HoldComplete[ BND$], free}
									],
									HoldComplete[ $TMHOR$AlphaEquivalentGround][
										Prepend[ Map[ {ToExpression[ prefix <> VarName[ #, Occurrence], InputForm, HoldComplete]}&, v],
												{ToExpression[ VarName[ First[ ai], Occurrence], InputForm, HoldComplete]}
										],
										{HoldComplete[ BND$], free}
									]
								]
							]
						], hovars],
					Null
				]
			];
			hovars = GatherBy[ Cases[ hovars[[All, 1]], _[ {_, _SEQ0$|_SEQ1$}, _, _, _]], First];
			cl = Join[ Map[ Function[ v,
					comp = ToExpression[ VarName[ First[ v], Len], InputForm, Hold];
					HoldComplete[ $TMHOR$SameLength][ comp, Map[ {ToExpression[ prefix <> VarName[ #, Occurrence], InputForm, HoldComplete]}&, v]]
				], hovars],
				cl
			]
		];
		If[ cl === {},
			new,
			With[ {n = new, cond = And@@cl},
				Condition[ n, cond]
			]
		]
	]
	
constructRangePattern[ rng_RNG$, noBlanks_List, prefix_String] :=
	$dummy@@MapThread[
		With[ {v2p = varToPattern[ #1, !MemberQ[ noBlanks, #1], prefix, LocalPattern -> True]},
			Switch[ #2,
				RNDRNG$[ _VAR$, VAR$[ _, _, _, _]],
				With[{ rp = Last[ varToPattern[ Last[ #2], False, prefix]]},
					Pattern@@Append[ rp, Blank[][ Last[ v2p], BlankNullSequence[]]]
				],
				_,
				#2 /. v2p
			]
		]&, {rngVariables[ rng], List@@rng}
	]
	
varToPattern[ v_, blank_, opts___?OptionQ] := varToPattern[ v, blank, "", opts]
varToPattern[ v_, False, Except[""], opts___?OptionQ] := varToPattern[ v, False, "", opts]
varToPattern[ v:VAR$[ {_, base_}], blank_, prefix_String, opts___?OptionQ] :=
	Module[ {lp = LocalPattern /. {opts} /. Options[ constructPattern]},
		With[ {new = ToExpression[ prefix <> VarName[ v], InputForm, HoldComplete]},
			With[ {rhs =
					If[ blank,
						If[ lp,
							Pattern@@Append[ new, VAR$[ {Blank[], Switch[ base, _SEQ0$, Blank[ SEQ0$], _SEQ1$, Blank[ SEQ1$], _Symbol, Blank[ Symbol]]}]],
							Pattern@@Append[ new, Blank[]]
						],
						new
					]
				},
				v :> rhs
			]
		]
	]
varToPattern[ v:VAR$[ {_, base_}, __], True, prefix_String, opts___?OptionQ] :=
	Module[ {new, useVar},
		useVar = Length[ v] < 4 && (SameFO /. {opts} /. Options[ constructPattern]);
		new = ToExpression[ prefix <> VarName[ v, If[ TrueQ[ useVar], Variable, Occurrence]], InputForm, HoldComplete];
		Switch[ base,
			_SEQ0$,
			With[ {rhs = Pattern@@Append[ new, BlankNullSequence[]]},
				v :> rhs
			],
			_SEQ1$,
			With[ {tmp1 = Pattern@@Append[ new, BlankSequence[]]},
				With[ {rhs = Condition[ tmp1, isNonEmptySequence[ new]]},
					v :> HoldPattern[ rhs]
				]
			],
			_,
			With[ {rhs = PatternTest[ Pattern@@Append[ new, Blank[]], isIndividual]},
				v :> rhs
			]
		]
	]
varToPattern[ v:VAR$[ _, __], False, prefix_String, opts___?OptionQ] :=
	Module[ {useVar},
		useVar = Length[ v] < 4 && (SameFO /. {opts} /. Options[ constructPattern]);
		With[ {new = ToExpression[ prefix <> VarName[ v, If[ TrueQ[ useVar], Variable, Occurrence]], InputForm, HoldComplete]},
			v :> new
		]
	]
varToPattern[ v:VAR$[ Except[ {_, _}]], _, _String, ___?OptionQ] := v -> v
	
(* The only occurrence of the 'i'-th first-order global variable is just that variable. *)
(* Attention! Function 'varName' with lower-case 'v' already exists in "FormulaManipulation.m". *)
VarName[ VAR$[ {i_, _}, -1, _], Occurrence] := "VAR$" <> ToString[ i]
VarName[ VAR$[ {i_, _}, $TMHOR$FO|$TMHOR$HO], Occurrence] := "VAR$" <> ToString[ i]
VarName[ VAR$[ {i_, _}, j_, ___], Occurrence] := "VAR$" <> ToString[ i] <> "$" <> ToString[ j]
VarName[ VAR$[ {i_, _}, __], Variable] := "VAR$" <> ToString[ i]
VarName[ VAR$[ {i_, _} ,_, _, _], ArgPositions] := "POS$" <> ToString[ i]
VarName[ VAR$[ {i_, _}, _, _, _], ArgPositions, a_] := "POS$" <> ToString[ i] <> "$" <> ToString[ a]
VarName[ {i_, _}, ArgPositions] := "POS$" <> ToString[ i]
VarName[ {i_, _}, ArgPositions, a_] := "POS$" <> ToString[ i] <> "$" <> ToString[ a]
VarName[ VAR$[ {i_, _}, j_, _, _], Bound] := "BND$" <> ToString[ i] <> "$" <> ToString[ j]
VarName[ VAR$[ {i_, _}, _, _, _], Len] := "LEN$" <> ToString[ i]
VarName[ {i_, _}, Len] := "LEN$" <> ToString[ i]
VarName[ VAR$[ {i_, _}, j_, $TMHOR$HO], Bound] := "BND$" <> ToString[ i] <> "$" <> ToString[ j]
VarName[ VAR$[ {i_, _}], ___] := "LOC$" <> ToString[ i]

(* 'bindingPositions[ expr ]' collects all local variables that are bound in 'expr'.
It returns a list of all those variables, together with the position of the range (SIMPRNG, SETRNG, ...) they are bound in. *)
bindingPositions[ expr_] :=
	Module[ {pos, tmp},
		pos = Position[ expr, _[ _RNG$, ___]];
		Map[ Function[ p,
				tmp = If[ p === {}, expr, Extract[ expr, p]];
				Sequence@@MapIndexed[ #1 -> Join[ p, {1, First[ #2]}]&, rngVariables[ First[ tmp]]]
			], pos
		]
	]
	



(* ::Section::Closed:: *)
(* Methods needed for Random Ranges *)


(* 'l' ... left-hand-side of the rewrite rule,
'r' ... right-hand-side of the rewrite rule,
'vars' ... list of all higher-order variables,
'glob' ... list of all global variables (first- and higher-order). *)
addRangeVariables[ l_, r_, vars_List, glob_List] :=
	Module[ {tmp, rc = {}, new = {}, rv, lhs, rhs, newCond, done},
		tmp = l //. q_[ rng_RNG$, c:(_VAR$[ ___VAR$]), rest___] :> q[ rng, And$TM[ c], rest]
				//. (q:Except[ _List])[ rng_RNG$, c:And$TM[ ___, _VAR$[ ___VAR$], ___], rest___] :> (
						rv = rngVariables[ rng];
						done = {};
						(* We employ a first come, first served strategy, meaning that the HO-predicates that come late
						in the conjunction can only be used as range-conditions for those variables for which no earlier
						HO-predicate can be used. *)
						newCond = Replace[ c,
								v_VAR$[ args___VAR$] /; (MemberQ[ vars, v] && Intersection[ {args}, glob] === {} && Length[ DeleteDuplicates[ {args}]] === Length[ {args}]) :>
									Module[ {depVars = {}, i = 1},
										Scan[ (If[ !MemberQ[ rv, #], AppendTo[ depVars, i]]; ++i)&, {args}];
										AppendTo[ rc, v -> DeleteCases[ MapThread[
																(i = Position[ {args}, #1, {1}, 1];
																If[ i === {},
																	Null,
																	i = First[ i];
																	depVars = Union[ depVars, i];
																	If[ MemberQ[ done, #1],
																		Null,
																		AppendTo[ done, #1];
																		Switch[ #2, _SIMPRNG$, First[i] -> depVars, _, Null]
																	]
																])&,
																{rv, List@@rng}],
																Null
															]
													];
										Null
									],
								{1}
							];
						{q, rng, newCond, rest}
					);
		
		(* 'rc' is a list consisting of rules of the form 'v -> {j1 -> {v1, dep1}, ...}',
		where every 'v' is a higher-order global variable appearing in 'l' only as condition of
		quantifiers only applied on distinct local variables.
		'ji' is the index among the arguments of 'v' of the 'i'-th simple-range variable bound by the quantifiers in whose scope 'v' appears;
			Bound variables that do not appear at all among the arguments of 'v' are omitted entirely,
			and the lists '{j1, ...}' have to be identical for all occurrences of 'v' (otherwise 'v' is not listed at all).
		'vi' is the name of the new range variable corresponding to the range of the 'i'-th simple-range variable bound by the quantifiers in whose scope 'v' appears, and
		'depi' is the list of indices of the arguments of 'v' 'vi' may depend upon. *)
		rc = DeleteCases[ Replace[ GatherBy[ rc, First],
								{
									{(v_ -> p:{__Rule})..} /; FreeQ[ tmp, v] :>
									Module[ {nv = Table[ $dummy[ {$varIndex++, Null}], {Length[ p]}]},
										new = Join[ new, nv];
										v -> MapThread[ First[ #1] -> {#2, Last[ #1]}&, {p, nv}]
									],
									_List -> Null
								},
								{1}
							],
							Null
						];
		
		lhs = l //. q_[ rng_RNG$, cond:(_VAR$[ ___]|And$TM[ ___, _VAR$[ ___], ___]), rest___] :>
					q@@Join[ makeRNDRNG[ rng, cond, rc, Left], {rest}] /. $dummy -> VAR$;
		
		rhs = r //. q_[ rng_RNG$, cond:(_VAR$[ ___]|And$TM[ ___, _VAR$[ ___], ___]), rest___] :>
					q@@Join[ makeRNDRNG[ rng, cond, rc, Right], {rest}]
				//. v_VAR$[ args___] :>
					Module[ {data, n, dep},
						(
							And$TM@@Append[
								Map[ Function[ d, {n, dep} = Last[ d]; Hold[ $TMHOR$RangeToCondition][ Part[ n[ args], dep]]], data],
								($dummy@@v)[ args]
							]
						) /; ListQ[ data = v /. rc]
					]
				/. $dummy -> VAR$;
		
		{lhs, rhs, new /. $dummy -> VAR$}
	]
	
makeRNDRNG[ rng_, v_VAR$[ args___], rc_List, Left] :=
	Module[ {data, rv, newRng = rng, newCond = ($dummy@@v)[ args], i, n, dep, j, argList = {args}, x},
		If[ ListQ[ data = (v /. rc)],
			rv = rngVariables[ rng];
			Scan[ Function[ d,
					i = First[ d];
					{n, dep} = Last[ d];
					x = argList[[i]];
					dep = Part[ argList, dep];
					j = Position[ rv, x, {1}, 1];
					newRng = ReplacePart[ newRng, j -> RNDRNG$[ x, n@@dep]]
				], data
			]
		];
		{newRng, newCond}
	]
makeRNDRNG[ rng_, v_VAR$[ args___], rc_List, Right] :=
	Module[ {data, rv, newCond = {}, newRng = rng, i, n, dep, j, argList = {args}, boundAfter, x},
		If[ ListQ[ data = (v /. rc)],
			rv = rngVariables[ rng];
			Scan[ Function[ d,
					i = First[ d];
					x = argList[[i]];
					{n, dep} = Last[ d];
					dep = Part[ argList, dep];
					j = Position[ rv, x, {1}, 1];
					If[ j === {} || (j = First[ First[ j]]; !MatchQ[ newRng[[j]], _SIMPRNG$]),
						AppendTo[ newCond, Hold[ $TMHOR$RangeToCondition][ n@@dep]],
						
						boundAfter = Drop[ rv, j];
						If[ FreeQ[ dep, Alternatives@@boundAfter],
							newRng = ReplacePart[ newRng, j -> RNDRNG$[ x, n@@dep]],
							AppendTo[ newCond, Hold[ $TMHOR$RangeToCondition][ n@@dep]]
						]
					]
				], data
			];
			Switch[ newCond,
				{},
				{newRng, ($dummy@@v)[ args]},
				_,
				{newRng, And$TM@@Append[ newCond, ($dummy@@v)[ args]]}
			],
			{rng, ($dummy@@v)[ args]}
		]
	]
makeRNDRNG[ rng_, cond_And$TM, rc_List, side_] :=
	Module[ {newRng = rng, newCond = cond, tmp},
		newCond = Map[ (tmp = makeRNDRNG[ newRng, #, rc, side]; newRng = First[ tmp]; Last[ tmp])&, cond];
		{newRng, newCond}
	]
makeRNDRNG[ rng_, cond_, _, _] := {rng, cond}

(* 'Unifiable[ {expr1, expr2} ]' checkes whether 'expr1' and 'expr2' can be unified (modulo alpha-equivalence).
All local variables are treated as constants and all global variables are treated as variables
(where applications of higher-order variables to arguments are of course replaced only by the variable);
More precisely, every first-order _variable_ and every higher-order _occurrence_ become variables
(alpha-equivalent occurrences of the same higher-order variable become the same unification variable, though),
but the unification variables in 'expr1' are made distinct from those in 'expr2'
(in the original expressions; this may change during the unification process)!
However, each higher-order occurrence 'o' may only be substituted by an expression containing only occurrences of
those local variables that appear free in the original arguments of 'o'.

The purpose of 'Unifiable' is to check whether two arguments of a higher-order occurrence 'o' can be instantiated
by the same expressions. If so, 'o' cannot be used for determining the positions of these arguments,
because it cannot distinguish between them (at least in some situations) - unless one of them is already known.

Actually, for all expressions 'e' and 'f' we require the following:
Let 'p' and 'q' be the patterns corresponding to 'e' and 'f', respectively, where 'p' and 'q' are generated by
'constructPattern'. If there is an expression 'expr' that is matched by one of 'p' or 'q' and also by a
(not necessarily proper) subexpression of the other, 'Unifiable[{e, f}]' must return 'True'.
However, even if no such expression exists, 'Unifiable' might still return 'True'. *)
Unifiable[ _[ l_, r_]] :=
	Module[ {fvConds = {}, e, f},
		e = l /. {VAR$[ i_, _, {_, j_}, _][ args___] :>
				(
					If[ !MemberQ[ fvConds, {j, Left} -> _List],
						AppendTo[ fvConds, {j, Left} -> Cases[ freeVariables[ {args}], VAR$[ _]]]
					];
					VAR$[ i, j, Left]
				),
				VAR$[ i_, _, _] :> VAR$[ i, Left]};
		f = r /. {VAR$[ i_, _, {_, j_}, _][ args___] :>
				(
					If[ !MemberQ[ fvConds, {j, Right} -> _List],
						AppendTo[ fvConds, {j, Right} -> Cases[ freeVariables[ {args}], VAR$[ _]]]
					];
					VAR$[ i, j, Right]
				),
				VAR$[ i_, _, _] :> VAR$[ i, Right]};
		
		(* If the unification problem involves sequence variables, we immediately return 'True'.
		This could be improved in the future. *)
		If[ !FreeQ[ {e, f}, VAR$[{ _, _SEQ0$|_SEQ1$},__]], Return[ True]];
		
		Unifiable[ e, f, fvConds, True] || Unifiable[ f, e, fvConds, False]
	]
Unifiable[ a:((f:Except[ VAR$])[ args___]), b_, fvConds_, compareWhole_] :=
	(* Note that 'a' might be a quantifier! *)
	(compareWhole && Unifiable[ {{a, b}}, fvConds]) || Unifiable[ f, b, fvConds, True] || Or@@Map[ Unifiable[ #, b, fvConds, True]&, Hold[ args]];
Unifiable[ a_, b_, fvConds_, compareWhole_] :=
	(compareWhole && Unifiable[ {{a, b}}, fvConds]);
Unifiable[ {}, _List] := True
Unifiable[ {{f_[ r:RNG$[ rng___], a___], g_[ s:RNG$[ sng___], b___]}, rest___}, fvConds_List] :=
	Module[ {rv = rngVariables[ r], sv = rngVariables[ s], repl},
		If[ Length[ Hold[ a]] === Length[ Hold[ b]] && Length[ rv] === Length[ sv],
			repl = MapThread[ If[ sequenceTypeEqual[ #1, #2], #1 -> #2, $Failed]&, {rv, sv}];
			If[ MemberQ[ repl, $Failed],
				False,
				Unifiable[ Join[ {{f, g}}, Transpose[ {{rng, a} /. repl, {sng, b}}], {rest}], fvConds /. repl]
			],
			False
		]
	]
Unifiable[ {{v:VAR$[ _, _], b_}, rest___}, fvConds_List] :=
	Module[ {fv = freeVariables[ b]},
		If[ MemberQ[ fv, VAR$[ _]|v],
			False,
			(* None of the higher-order variables appearing in 'b' may be instantiated with an expression containing
			a local variable, because then also the instance of 'v' would contain a local variable. *)
			Unifiable[ {rest} /. v -> b, DeleteCases[ fvConds, ({j_, p_} -> _List) /; MemberQ[ fv, VAR$[ _, j, p]]]]
		]
	]
Unifiable[ {{v:VAR$[ _, j_, p_], b_}, rest___}, fvConds_List] :=
	Module[ {fv = freeVariables[ b], vars = {j, p} /. fvConds /. {j, p} -> {}},
		If[ MemberQ[ fv, VAR$[ _]?(!MemberQ[ vars, #]&)|v],
			False,
			Unifiable[ {rest} /. v -> b, Replace[ fvConds, ({j0_, p0_} -> l_List) /; MemberQ[ fv, VAR$[_, j0, p0]] :> ({j0, p0} -> Intersection[ l, vars]), {1}]]
		]
	]
Unifiable[ {{a_, v:VAR$[ _, __]}, rest___}, fvConds_List] := Unifiable[ {{v, a}, rest}, fvConds]
Unifiable[ {{(f:Except[ VAR$])[ a___], (g:Except[ VAR$])[ b___]}, rest___}, fvConds_List] :=
	If[ Length[ Hold[ a]] === Length[ Hold[ b]],
		Unifiable[ Join[ Transpose[ {{f, a}, {g, b}}], {rest}], fvConds],
		False
	]
Unifiable[ {{a_, b_}, rest___}, fvConds_List] :=
	If[ a === b,
		Unifiable[ {rest}, fvConds],
		False
	]
	
(* 'makeShapeAlphaConditions[ expr, hoVars ]' returns 'allOccurrences' and 'correctShape' conditions that have
to be fulfilled by the higher-order variables 'hoVars' occuring in 'expr'. *)
makeShapeConditions[ expr_, hoVars_List, locStruct_LOCSTRUCT$, signalFailure_:False] :=
	Module[ {tmp, gen = Cases[ hoVars, VAR$[ _, _, _, True]], out = {}, l, arg, vari, data, rest,
			unknownArgs, genCascade, occ, siblings, ai, f},
	
		genCascade = {gen};
		
		(* Construct a list
		'{i -> {1 -> {{{j, gen0}, {a1, a2, ...}|True|False, patt, {f1, ...}, {s1, ...}, {v1, ...}}, ...}, 2 -> {...}, ...}, ...}'
		with the following meaning:
		'i' are the indices of all the higher-order variables in 'allVars'.
		'j' are the indices of all the occurrences of 'i'.
		'gen0' specifies whether 'j' is instantiated in the main pattern.
		'patt' is a pattern expression corresponding to the 1st argument of 'j'.
		The 'fi' are occurrences of global first-order variables that are generated when the 1st argument is
			determined/checked by 'j' (one occurrence of each global first-order variable).
		The 'si' are those higher-order sequence variables where at least one occurrence is generated when the
			1st argument is determined/checked by 'j'.
		The 'vi' are the higher-order global and local variables that are generated when the 1st argument is
			determined/checked by 'j'.
		The 'ai' correspond to the arguments of 'i' that must be known already such that the positions of the
			1st argument of 'i' can be computed from 'j' (i.e. those arguments that are unifiable with the 1st one);
			If this is not possible at all, the whole object is 'False';
			If the 1st argument of 'i' has already been computed from a sibling of 'j', the whole object is 'True'.
			The distinction between 'True' and 'False' is only for better readability, but it does not influence the behaviour of the program. *)
		l = Replace[ GatherBy[ Cases[ expr, (v:VAR$[ i_, j_, {arity_, _}, gen0_])[ args___] /; (Length[ {args}] === arity && MemberQ[ hoVars, v]) :>
										(i -> {{j, gen0} -> {args}}), {0, Infinity}, Heads -> True
									],
								First
							],
					elem:{(i_ -> _List)..} :>
						(i -> Replace[ Transpose[ Map[ Function[ x, Map[ (First[ #] -> Prepend[ Last[ #], First[ x]])&, makeShapeConditions[ Last[ x], locStruct]]],
													   Flatten[ elem[[All, 2]], 1]
													 ]
												],
										e:{(a_ -> _List)..} :> (a -> e[[All, 2]]), {1}]
						),
					{1}
				];
		
		If[ signalFailure,
			(* We collect all variables 'v' having an argument index 'a' such that every
			occurrence 'o' of 'v' has 'False' as the second element of the respective list in 'l'.
			If the list 'f' of these variables is non-empty, '{$Failed, f}' is returned. *)
			f = Cases[ l, (v_ -> {___, _ -> {{_, False, ___}...}, ___}) :> v];
			If[ f =!= {},
				out = {$Failed, f};
				l = {}
			]
		];
		
		While[ l =!= {},
			If[ !Catch[
				Scan[ Function[ var,
					vari = First[ var];
					unknownArgs = Cases[ Last[ var], HoldPattern[ a_ -> {___, {_List, _List, _RuleDelayed, _List, _List, _List}, ___}] :> a];
					rest = $Failed;
					arg = Cases[ Last[ var],
								HoldPattern[ a_ -> {p0___, d:{{_?(MemberQ[ gen, _[ vari, #, ___]]&), _}, ka_List, _RuleDelayed, _List, _List, _List}, p1___}] /; (Intersection[ ka, unknownArgs] === {}) :>
									{a, d, {p0, p1}},
								{1},
								1
							];
					
					If[ arg =!= {},
						{arg, data, rest} = First[ arg];
						AppendTo[ out, makeAllOccurrences[ vari, arg, data]],
						
						arg = Cases[ Last[ var],
									HoldPattern[ a_ -> {p0:({_List, True|False, _RuleDelayed, _List, _List, _List}...), d:{{_?(MemberQ[ gen, _[ vari, #, ___]]&), _}, True|False, _RuleDelayed, _List, _List, _List}, p1:({_List, True|False, _RuleDelayed, _List, _List, _List}...)}] :>
										{a, d, {p0, p1}},
									{1}, 1];
						If[ arg =!= {},
							{arg, data, rest} = First[ arg];
							AppendTo[ out, makeCorrectShape[ vari, arg, data]]
						]
					];
					
					If[ ListQ[ rest],
						occ = data[[1, 1]];
						{ai, siblings} = First[ Cases[ genCascade, {pre___, _[ vari, occ, {___, ai0_}, _], post___} :>
																	{ai0, {pre, post}},
														{1}, 1]
										];
						If[ Last[ data] === {},
							(* Only first-order global variables are generated by this occurrence.
							This means that we can entirely remove all alpha-equivalent siblings of 'occ' from 'rest', instead of replacing them. *)
							tmp = Cases[ siblings, _[ _, j_, {___, ai}, _] :> j];
							siblings = DeleteCases[ siblings, _[ _, _, {___, ai}, _]];
							If[ tmp =!= {},
								(* Delete all '{{i, _}, __}' from 'rest', for every 'i' in 'tmp'. *)
								rest = DeleteCases[ rest, {{_?(MemberQ[ tmp, #]&), _}, __}]
							],
							
							AppendTo[ genCascade, Cases[ Last[ data], VAR$[ _, _, _, _]]];
							gen = Join[ gen, Last[ genCascade]]
						];
						tmp = Cases[ siblings, _[ _, j_, _, _] :> j];
						If[ tmp =!= {},
							(* Replace all '{{i, _}, _, __}' in 'rest' by '{{i, _}, True, __}', for every 'i' in 'tmp'.
							Note that this happens independently of arguments being alpha-equivalent! *)
							rest = Replace[ rest, {x:{_?(MemberQ[ tmp, #]&), _}, _, post__} :> {x, True, post}, {1}]
						];
						
						tmp = If[ rest === {},
							DeleteCases[ Last[ var], arg -> _List],
							Replace[ Last[ var], (arg -> _List) -> (arg -> rest), {1}]
						];
						l = If[ tmp === {},
							DeleteCases[ l, vari -> _List],
							Replace[ l, (vari -> _List) -> (vari -> tmp), {1}]
						];
						Throw[ True]
					]
				], l];
				Throw[ False]
			],
			
			(* We try to remove only those variables that do not have occurrences with pairwise-ununifiable
			arguments. *)
			rest = DeleteCases[ l, _ -> {(_ -> {___, {_, {}, ___}, ___})..}];
			If[ rest === {}, rest = l];
			out = {$Failed, rest[[All, 1]]};
			l = {};
			]
		];
		out
	]
makeShapeConditions[ args_List, locStruct_LOCSTRUCT$] :=
	Module[ {newArgs, fv, gv, flArgs, ran = Range[ Length[ args]], ka, unif, patt, gen},
		newArgs = args /. (v:VAR$[ _, _, _, _])[ ___] :> v;
		fv = Map[ freeVariables, newArgs];
		flArgs = Flatten[ Position[ fv, _List?(MemberQ[ #, VAR$[ _]]&), {1}]];
		unif = Cases[ Map[ {#, Unifiable[ {args[[First[ #]]], args[[Last[ #]]]}]}&,
							Select[ Tuples[ ran, 2], (First[ #] < Last[ #] && MemberQ[ flArgs, First[ #]|Last[ #]])&]
					  ],
					  {p_, True} :> p
				];
		
		(* This is one place where code has to be changed if instead of alpha-equivalent instances of global variables we want identical ones.
		Just add option 'SameFO -> True' to all calls of 'constructPattern' below. *)
		MapThread[ Function[ {a, free, index},
			ka = If[ MemberQ[ flArgs, index],
				Cases[ unif, ({index, o_}|{o_, index}) :> o],
				False
			];
			gv = Cases[ free, _[ _, __]];
			
			(* We need the positions of the local variables bound in 'a' in any case. *)
			patt = constructPattern[ a, Cases[ free, _[ _]], "$", locStruct, SameFO -> False];
			gen = Join[ gv, Flatten[ Cases[ a, r_RNG$ :> rngVariables[ r], {0, Infinity}, Heads -> True], 1]];
			
			(* We only need one occurrence of each global first-order variable in 'gen'. *)
			gen = {DeleteDuplicates[ Cases[ gen, _[ _, _, _]], First[ #1] === First[ #2]&], Cases[ gen, _[ _]|_[ _, _, _, _]]};
			gen = Insert[ gen, DeleteDuplicates[ Cases[ Last[ gen], _[{_, _SEQ0$|_SEQ1$}, _, _, _]], First[ #1] === First[ #2]&], 2];
			
			(* This is one place where code has to be changed if instead of alpha-equivalent instances of global variables we want identical ones.
			Just return 'VarName[#, Variable]' for all first-order variables. *)
			With[ {rhs = {Map[ {ToExpression[ "$" <> VarName[ #, Occurrence], InputForm, Hold]}&, First[ gen]],
						  Map[ Hold[ Length][ {ToExpression[ "$" <> VarName[ #, Occurrence], InputForm, Hold]}]&, gen[[2]]],
						  Map[ With[ {n = ToExpression[ "$" <> VarName[ #, Occurrence], InputForm, Hold]}, Switch[ #, VAR$[ _], n, _, {n}]]&, Last[ gen]]}},
				index -> Join[ {ka, patt :> rhs}, gen]
			]
		], {newArgs, fv, ran}]
	]
	
makeAllOccurrences[ i_, a_, {{j_, gen0_}, knownArgs_List, patt_, {}, {}, {}}] :=
	Module[ {var = VAR$[ i, j, 0, True]},
		With[ {ps = ToExpression[ VarName[ i, ArgPositions, a], InputForm, Hold],
			   occ = With[ {o = ToExpression[ VarName[ var, Occurrence], InputForm, Hold]}, If[ gen0, {o}, o]],
			   ka = Map[ ToExpression[ VarName[ i, ArgPositions, #], InputForm, Hold]&, knownArgs]},
			Hold[ $TMHOR$AllOccurrences[ ps, occ, patt, ka] =!= $Failed]
		]
	]
makeAllOccurrences[ i_, a_, {{j_, gen0_}, knownArgs_List, patt_, fo_List, seq_List, vars_List}] :=
	Module[ {var = VAR$[ i, j, 0, True]},
		With[{ ps = ToExpression[ VarName[ i, ArgPositions, a], InputForm, Hold],
			   occ = With[ {o = ToExpression[ VarName[ var, Occurrence], InputForm, Hold]}, If[ gen0, {o}, o]],
			   ka = Map[ ToExpression[ VarName[ i, ArgPositions, #], InputForm, Hold]&, knownArgs],
			   nfo = Map[ ToExpression[ VarName[ #, Variable], InputForm, Hold]&, fo],
			   len = Map[ ToExpression[ VarName[ #, Len], InputForm, Hold]&, seq],
			   nvars = Map[ ToExpression[ VarName[ #, Occurrence], InputForm, Hold]&, vars]},
			Hold[ $TMHOR$Update[ nfo, len, nvars, $TMHOR$AllOccurrences[ ps, occ, patt, ka]]]
		]
	]
	
makeCorrectShape[ i_, a_, {{j_, gen0_}, _, patt_, {}, {}, {}}] :=
	Module[ {var = VAR$[ i, j, 0, True]},
		With[ {ps = ToExpression[ VarName[ i, ArgPositions, a], InputForm, Hold],
			   occ = With[ {o = ToExpression[ VarName[ var, Occurrence], InputForm, Hold]}, If[ gen0, {o}, o]]},
			Hold[ $TMHOR$CorrectShape[ ps, occ, patt] =!= $Failed]
		]
	]
makeCorrectShape[ i_, a_, {{j_, gen0_}, _, patt_, fo_List, seq_List, vars_List}] :=
	Module[ {var = VAR$[ i, j, 0, True]},
		With[ {ps = ToExpression[ VarName[ i, ArgPositions, a], InputForm, Hold],
			   occ = With[ {o = ToExpression[ VarName[ var, Occurrence], InputForm, Hold]}, If[gen0, {o}, o]],
			   nfo = Map[ ToExpression[ VarName[ #, Variable], InputForm, Hold]&, fo],
			   len = Map[ ToExpression[ VarName[ #, Len], InputForm, Hold]&, seq],
			   nvars = Map[ ToExpression[ VarName[ #, Occurrence], InputForm, Hold]&, vars]},
			Hold[ $TMHOR$Update[ nfo, len, nvars, $TMHOR$CorrectShape[ ps, occ, patt]]]
		]
	]
	
makeAlphaEquivalentFree[ locStruct_LOCSTRUCT$, hoVars_List] :=
	Module[ {ai, loc, classes,
			ls = Cases[ locStruct, HoldPattern[ o_ -> l_List] /; MemberQ[ hoVars, o] :> (o -> Complement[ First[ l], Last[ l]])]},
		DeleteCases[ Replace[ GatherBy[ ls, #[[1, 1]]&],
				all:{(VAR$[ i_, _, {arity_, _}, _] -> _List)..} :> (
					(* It is not sufficient to take only one representative from each alpha-equivalence class,
					although it has already been checked that the instances of the occurrences belonging to the same equivalence class are alpha-equivalent:
					Not all of them might be instantiated. *)
					classes = GatherBy[ all, #[[1, 3, 2]]&];
					loc = Union@@all[[All, 2]];
					If[ Length[ classes] === 1 && (loc === {} || ((ai = classes[[1, 1, 1, 3, 2]]; And@@Map[ Negative, ai]) && Length[ DeleteDuplicates[ ai]] === Length[ ai])),
						(* There is only one equivalence class, and either all the arguments are pairwise distinct local variables,
						or there are no freeness conditions. Hence, no condition needs to be generated. *)
						Null,
						
						With[ {ps = Table[ ToExpression[ VarName[ i, ArgPositions, a], InputForm, Hold], {a, arity}],
							   occ = Map[ With[ {n = ToExpression[ VarName[ First[ #], Occurrence], InputForm, Hold]}, If[ Last[ First[ #]], {n}, n]]&, classes, {2}]},
							   loc = Map[ ToExpression[ VarName[ #], InputForm, Hold]&, loc];
							Hold[ $TMHOR$AlphaEquivalentFreeHO][ Sequence@@occ, ps, loc]
						]
					]
				),
				{1}],
			Null
		]
	]
	
	


(* ::Section::Closed:: *)
(* Runtime Methods *)


(* All runtime methods have the prefix "$TMHOR$". *)

(* General remark on program variables:
Program variables standing for global variables can be instantiated by both individual expressions and sequences
of expressions, depending on whether the corresponding variable in the original formula is a sequence variable or
not. Hence, since plain sequences of expressions may cause some problems during evaluation, (almost) all such
program variables are in fact instantiated by lists, even if they actually correspond to individual variables in
the formula. The lists are removed just before the new expression is constructed (using function "makeSequence").
More precisely, this holds for program variables corresponding to first-order variables ("VAR$i"), and for program
variables corresponding to higher-order occurrences that do not occur in the main pattern ("VAR$i$j"; Pattern
variables are the only exceptions, since we cannot change what they are instantiated with).
This remark does not apply for local variables, as they have to be individual expressions anyway (even if they are
sequence variables).
*)

(* 'allOccurrences[ pos\[UpArrow], expr:{expr1, ..., exprE}, patt :> {{..., genG}, {..., genL}, {..., genH}}, {p1, ...} ]'
does the following: If every 'expri' is '$TMHOR$Unknown' or 'E' is 0, it immediately returns a triple consisting
of a list of 'G' copies of '{$TMHOR$Unknown}', a list of 'L' copies of '-1', and a list of 'H' copies of
'{$TMHOR$Unknown}' and leaves 'pos' unchanged.
Otherwise, if none of the 'expri' is '$TMHOR$Unknown', it first removes all the 'pi' that are no list and then
distinguishes two cases, depending on whether 'pos' is a list or not:

- If 'pos' is a symbol, then the function first collects all positions 'a' of subexpressions in 'expr' matching
	'patt', and removes those positions that are either sub- or superpositions of any of the remaining 'pi'.
	If 'a' is empty, the function again returns a triple consisting of a list of 'G' copies of '{$TMHOR$Unknown}',
	a list of 'L' copies of '-1', and a list of 'H' copies of '{$TMHOR$Unknown}'; Otherwise, all subexpressions
	of 'expr' at the positions in 'a' are collected. If they are all alpha-equivalent to each other, then
	'Replace[inst, patt :> {gen1, ...}]' is returned, where 'inst' is one of the alpha-equivalent subexpressions.
	Otherwise, '$Failed' is returned.
	In any case, 'pos' is set to 'a'.
- If 'pos' is already a list, the function does essentially the same as if it was no list, only that the list of
	positions 'a' has to agree with 'pos'. If this is not the case, '$Failed' is returned immediately. 'pos' itself is left unchanged.

In all other cases, the function returns '$Failed' (This, however, merely indicates an internal error).
It is crucial that those 'pi' that are no lists are not assigned lists later. This is taken care of by the code
generation (makeShapeConditions). Also, note that the elements of 'pos' refer to positions IN THE LIST, i.e. no
position will ever be '{}'. *)
$TMHOR$AllOccurrences[ _, {$TMHOR$Unknown..}, _ :> {{g1___}, {g2___}, {g3___}}, _List] :=
	{Table[ {$TMHOR$Unknown}, {Length[ Hold[ g1]]}], Table[ -1, {Length[ Hold[ g2]]}], Table[ {$TMHOR$Unknown}, {Length[ Hold[ g3]]}]}
$TMHOR$AllOccurrences[ pos_, expr_List, patt_, argpos_List] :=
	Module[ {p, inst},
		{p, inst} = $TMHOR$AllOccurrences[ expr, patt, Flatten[ Cases[ argpos, _List], 1]];
		Switch[ pos,
			_Symbol,
			pos = p;
			inst,
			p,
			inst,
			_,
			$Failed
		]
	]
$TMHOR$AllOccurrences[ expr_List, rule:(patt_ :> {{g1___}, {g2___}, {g3___}}), argpos_List] :=
	Module[ {p, sub, invalid},
		p = Position[ ReplacePart[ expr, argpos -> $TMHOR$Unknown], patt];
		sub = $TMHOR$Extract[ expr, p];
		invalid = Position[ sub, _?(!FreeQ[ #, $TMHOR$Unknown]&), {1}];
		sub = Delete[ sub, invalid];
		p = Delete[ p, invalid];
		Switch[ sub,
			{},
			{{}, {Table[ {$TMHOR$Unknown}, {Length[ Hold[ g1]]}], Table[ -1, {Length[ Hold[ g2]]}], Table[ {$TMHOR$Unknown}, {Length[ Hold[ g3]]}]}},
			{__},
			With[ {f = First[ sub], r= Hold@@Rest[ sub]},
				If[ And@@Map[ $TMHOR$AlphaEquivalent[ f, #, {}]&, r],
					If[ Hold[ g1] === Hold[ g2] === Hold[ g3] === Hold[ ],
						{p, {{}, {}, {}}},
						{p, Replace[ f, rule]}
					],
					{$Failed, $Failed}
				]
			]
		]
	]
	
(* 'correctShape[ pos, expr:{expr1, ...}, patt :> {{..., genG}, {..., genL}, {..., genH}} ]'
immediately returns a triple consisiting of a list of 'G' copies of '{$TMHOR$Unknown}', a list of 'L' copies
of '-1', and a list of 'H' copies of '{$TMHOR$Unknown}' if every 'expri' is '$TMHOR$Unknown' or 'pos' is no list.
Otherwise, the function checks whether all elements of 'pos' are valid positions in 'expr' and whether all
subexpressions at those positions are alpha-equivalent instances of 'patt'. If so, either
'Replace[inst, patt :> {gen1, ...}]' for such an instance 'inst', or again a triple of a list of 'G' resp. 'H'
copies of '{$TMHOR$Unknown}' and a list of 'L' copies of '-1' is returned (depending on whether 'pos' is empty or
not). Otherwise, '$Failed' is returned.
It is crucial that those 'pi' that are no lists are not assigned lists later. This is taken care of by the code
generation (makeShapeConditions). *)
$TMHOR$CorrectShape[ _, {$TMHOR$Unknown..}, _ :> {{g1___}, {g2___} ,{g3___}}] :=
	{Table[ {$TMHOR$Unknown}, {Length[ Hold[ g1]]}], Table[ -1, {Length[ Hold[ g2]]}], Table[ {$TMHOR$Unknown}, {Length[ Hold[ g3]]}]}
$TMHOR$CorrectShape[ Except[ _List], _, _ :> {{g1___}, {g2___}, {g3___}}] :=
	{Table[ {$TMHOR$Unknown}, {Length[ Hold[ g1]]}], Table[ -1, {Length[ Hold[ g2]]}], Table[ {$TMHOR$Unknown}, {Length[ Hold[ g3]]}]}
$TMHOR$CorrectShape[ {}, _, _ :> {{g1___}, {g2___}, {g3___}}] :=
	{Table[ {$TMHOR$Unknown}, {Length[ Hold[ g1]]}], Table[ -1, {Length[ Hold[ g2]]}], Table[ {$TMHOR$Unknown}, {Length[ Hold[ g3]]}]}
$TMHOR$CorrectShape[ pos_List, expr_List, rule:(patt_ :> {{g1___}, {g2___}, {g3___}})] :=
	With[ {sub = Quiet[ Check[ $TMHOR$Extract[ expr, pos], $Failed]]},
		Switch[ sub,
			{__},
			With[ {f = First[ sub], r = Hold @@ Rest[ sub]},
				If[ MatchQ[ f, patt] && And @@ Map[ $TMHOR$AlphaEquivalent[ f, #, {}]&, r],
					If[ Hold[ g1] === Hold[ g2] === Hold[ g3] === Hold[ ],
						{{}, {}},
						Replace[ f, rule]
					],
					$Failed
				]
			],
			_,
			$Failed
		]
	]
	
$TMHOR$SameLength[ args_List] :=
	SameQ @@ Map[ Length, DeleteCases[ args, {$TMHOR$Unknown..}|Except[ _List]]]
$TMHOR$SameLength[ len_Integer, args_List] :=
	SameQ @@ Prepend[ Map[ Length, DeleteCases[ args, {$TMHOR$Unknown..}|Except[ _List]]], len]
$TMHOR$SameLength[ Except[ _Integer], args_List] :=
	$TMHOR$SameLength[ args]
	
(* 'update[ al:{a1, ...}, cl:{c1, ...}, bl:{b1, ...}, {ul:{u1, ...}, ll:{l1, ...}, vl:{v1, ...}}|$Failed ]'
immediately returns 'False' if the second argument is '$Failed' or the lists 'al' and 'ul' or the lists
'cl' and 'll' or the lists 'bl' and 'vl' are not of the same length. Otherwise, for each pair 'a', 'u' taken from
the respective lists, it does the following:
If 'a' is a list or 'u' is not a list or 'u' is a list entirely made up of '$TMHOR$Unknown', nothing happens.
Otherwise, it assigns 'u' to 'a'.
For each pair 'c' and 'l' taken from the respective lists, it does the following:
If 'c' is no symbol or 'l' is not a non-negative integer, nothing happens. Otherwise, it assigns 'l' to 'c'.
For each pair 'b' and 'v' taken from the respective lists, it assigns 'v' to 'b', regardless of their values.
In any case, the function returns 'True'. *)
$TMHOR$Update[ fo_List, seq_List, vars:{___Symbol}, {fov_List, len_List, varsv_List}] /; Length[ fo] === Length[fov] && Length[ seq] === Length[ len] && Length[ vars] === Length[ varsv] :=
	Module[ {del},
		MapThread[ Set, {vars, varsv}];
		del = DeleteDuplicates[ Join[ Position[ fo, Except[ _Symbol], {1}, Heads -> False],
									  Position[ fov, Except[ _List]|{$TMHOR$Unknown..}, {1}, Heads -> False]
								]
			];
		MapThread[ Set, {Delete[ fo, del], Delete[ fov, del]}];
		del = DeleteDuplicates[ Join[ Position[ seq, Except[ _Symbol], {1}, Heads -> False],
									  Position[ len, Except[ _Integer?NonNegative], {1}, Heads -> False]
								]
			];
		MapThread[ Set, {Delete[ seq, del], Delete[ len, del]}];
		True
	]
$TMHOR$Update[ ___] := False

(* 'alphaEquivalentFree[ {expr1, ...}, {var1, ...} ]' first immediately removes all of the 'expri' that are no
lists or lists where every element is '$TMHOR$Unknown'. The function returns 'True' or 'False' depending on whether
the remaining expressions are alpha-equivalent to each other _and_ none of the 'vari' occurs in the expressions
('freeOf').
'alphaEquivalentFree[ {e1, ...}, {f1, ...}, vars ]' first removes all the 'ei' and 'fi' that are no lists or lists
entirely made up of '$TMHOR$Unknown'. If no 'fi' remains, the function immediately returns 'True'. If no 'ei'
remains, 'alphaEquivalentFree[ {f1, ..}, vars ]' is returned; Otherwise, if 'e1' is the first of the remaining
'ei', 'alphaEquivalentFree[ {e1, f1, ...}, vars ]' is returned. *)
$TMHOR$AlphaEquivalentFree[ exprs_List, vars_List] :=
	With[ {e = DeleteCases[ exprs, {$TMHOR$Unknown..}|Except[ _List]]},
		Switch[ e,
			{},
			True,
			_,
			With[ {f = First[ e]},
				$TMHOR$FreeOf[ f, vars] &&
				Catch[
					Scan[ If[ $TMHOR$AlphaEquivalent[ f, #, {}] === False, Throw[ False]]&, Hold @@ Rest[ e]];
					Throw[ True]
				]
			]
		]
	]
$TMHOR$AlphaEquivalentFree[ exprs1_List, exprs2_List, vars_List] :=
	With[ {e1 = DeleteCases[ exprs1, {$TMHOR$Unknown..}|Except[ _List]],
			 e2 = DeleteCases[ exprs2, {$TMHOR$Unknown..}|Except[ _List]]},
		Switch[ e2,
			{},
			True,
			_,
			Function[ arg,
				With[ {f = First[ arg], r = Last[ arg]},
					$TMHOR$FreeOf[ f, vars] &&
					Catch[
						Scan[ If[ $TMHOR$AlphaEquivalent[ f, #, {}] === False, Throw[ False]]&, r];
						Throw[ True]
					]
				]
			][
				Switch[ e1,
					{},
					{First[ e2], Hold@@Rest[ e2]},
					_,
					{First[ e1], Hold@@e2}
				]
			]
		]
	]
	
(* In contrast to 'alphaEquivalentFree[ exprs, vars]', 'alphaEquivalentGround[ exprs, bound]' checks
	whether none of the elements of 'bound' occurs free in 'exprs'. *)
$TMHOR$AlphaEquivalentGround[ exprs_List, bound_] :=
	With[ {e = DeleteCases[ exprs, {$TMHOR$Unknown..}|Except[ _List]]},
		Switch[ e,
			{},
			True,
			_,
			With[ {f = First[ e]},
				$TMHOR$IsGround[ f, bound] &&
				Catch[
					Scan[ If[ $TMHOR$AlphaEquivalent[ f, #, {}] === False, Throw[ False]]&, Hold@@Rest[ e]];
					Throw[ True]
				]
			]
		]
	]
$TMHOR$AlphaEquivalentGround[ exprs1_List, exprs2_List, bound_] :=
	With[ {e1 = DeleteCases[ exprs1, {$TMHOR$Unknown..}|Except[ _List]],
			 e2 = DeleteCases[ exprs2, {$TMHOR$Unknown..}|Except[ _List]]},
		Switch[ e2,
			{},
			True,
			_,
			Function[ arg,
				With[ {f = First[ arg], r = Last[ arg]},
					$TMHOR$IsGround[ f, bound] &&
					Catch[
						Scan[ If[ $TMHOR$AlphaEquivalent[ f, #, {}] === False, Throw[ False]]&, r];
						Throw[ True]
					]
				]
			][
				Switch[ e1,
					{},
					{First[ e2], Hold@@Rest[ e2]},
					_,
					{First[ e1], Hold@@e2}
				]
			]
		]
	]
	
(* 'alphaEquivalentFreeHO[ exprs1:{e11, ...}, exprs2:{e21, ...}, ..., {pos1, ...}, {var1, ...} ]'
first immediately removes all of the 'eij' that are no lists or lists where every element is '$TMHOR$Unknown'.
Then all the 'posi' that are no lists and all the 'exprsi' that are empty are removed, and for each non-empty
'exprsi' only the first element 'ei' is kept. Afterwards, all subexpressions in all the 'ei' at positions in the
remaining 'posi' are replaced by 'Null'. Finally, the function returns 'True' or 'False' depending on whether the
resulting expressions are alpha-equivalent to each other _and_ none of the 'vari' occurs in the expressions ('freeOf'). *)
$TMHOR$AlphaEquivalentFreeHO[ exprs___List, pos_List, vars_List] :=
	Module[ {tmp, e, p = Flatten[ Cases[ pos, _List], 1]},
		e = DeleteCases[ Map[ (
								tmp = DeleteCases[ #, {$TMHOR$Unknown..}|Except[ _List]];
								Switch[	tmp,
									{},
									$TMHOR$Unknown,
									_,
									First[ tmp]]
							  )&,
							  {exprs}
						],
						$TMHOR$Unknown
			];
		Switch[ e,
			{},
			True,
			_,
			If[ p =!= {},
				e = Map[ ReplacePart[ #, p -> Null]&, e]
			];
			With[ {f = First[ e]},
				$TMHOR$FreeOf[ f, vars] &&
				Catch[
					Scan[ If[ $TMHOR$AlphaEquivalent[ f, #, {}] === False, Throw[ False]]&, Hold @@ Rest[ e]];
					Throw[ True]
				]
			]
		]
	]
	
$TMHOR$AlphaEquivalent[ a_Hold, b_Hold, repl_] := alphaEquivalent[ a, b, repl]
$TMHOR$AlphaEquivalent[ a_Hold, b_, repl_] := alphaEquivalent[ a, Hold[ b], repl]
$TMHOR$AlphaEquivalent[ a_, b_Hold, repl_] := alphaEquivalent[ Hold[ a], b, repl]
$TMHOR$AlphaEquivalent[ a_, b_, repl_] := alphaEquivalent[ Hold[ a], Hold[ b], repl]

$TMHOR$MakeSequence[ {a___}] := a
$TMHOR$MakeSequence[ _] := $TMHOR$Unknown

(* 'freeOf[ expr, vars ]' checks whether none of the members of 'vars' appears in 'expr'
(FreeQ; This also applies if 'expr' is/contains '$TMHOR$Unknown').
Before that, every item of 'vars' that does not have head 'Theorema`Language`VAR$' is dropped. *)
$TMHOR$FreeOf[ expr_, vars:{___, Except[ _VAR$], ___}] :=
	$TMHOR$FreeOf[ expr, Cases[ vars, _VAR$]]
$TMHOR$FreeOf[ expr_, {vars__}] := FreeQ[ expr, Alternatives[ vars]]
$TMHOR$FreeOf[ expr_, {}] := True

(* 'isGround[ expr, bound, free]' checks whether all of the elements of 'bound' that appear free in 'expr'
	are also included in 'free'.
	The function relies on distinct variables having different names, which is ensured by preprocessing
	the expression to be rewritten. *)
$TMHOR$IsGround[ expr_, {bound_List, free_List}] :=
	$TMHOR$IsGround[ expr, bound, free]
$TMHOR$IsGround[ expr_, bound_List] :=
	$TMHOR$IsGround[ expr, bound, {}]
$TMHOR$IsGround[ expr_, bound_List, free_List] :=
	Catch[
		Scan[
			If[ !MemberQ[ free, #],
				With[ {pos = Position[ expr, #]},
					If[ pos =!= {} && !MemberQ[ pos, {pre__, _, 1} /; Extract[ expr, {pre, 0}, Hold] === Hold[ RNG$]],
						Throw[ False]
					]
				]
			]&,
			bound
		];
		Throw[ True]	
	]

(* 'getVar[ {{expr1, ex1}, ...}, ex, vars, orig ]' returns a variable: If 'vars' contains an element 'v' such that
for each i either 'exi' contains 'v' or 'expri' is free of any variable with the same base name as 'v' (FreeQ),
and 'ex' does not contain any such variable either, then the _first_ such 'v' is returned.
Otherwise, 'freshVar[ {{expr1, ex1}, ...}, ex, orig ]' is returned. *)
$TMHOR$GetVar[ l_List, ex_List, {}, orig_] := $TMHOR$FreshVar[ l, ex, orig]
$TMHOR$GetVar[ l:{exprs___}, ex_List, {v:VAR$[ {_, n_}], rest___}, orig_] :=
	If[ !MemberQ[ ex, VAR$[ {_, n}]] && And @@ Map[ (MemberQ[ Last[ #], VAR$[ {_, n}]]||FreeQ[ First[ #], VAR$[ {_, n}]|VAR$[ n]])&, Hold[ exprs]],
		v,
		$TMHOR$GetVar[ l, ex, {rest}, orig]
	]
	
(* 'freshVar[ {expr1, ...}, ex, orig ]' returns the fresh variable 'origi' not appearing in any of the 'expri'
(FreeQ), and not included in the list of variables 'ex'. If 'orig' is a sequence variable, then the output is a
sequence variable as well. *)
$TMHOR$FreshVar[ l_List, ex_List, SEQ0$[ orig_Symbol]] :=
	VAR$[ {$varIndex++, SEQ0$[ $TMHOR$FreshName[ orig, Cases[ {l, ex}, (VAR$[ SEQ0$[ s_Symbol]]|VAR$[ {_, SEQ0$[ s_Symbol]}]) :> s, {0, Infinity}, Heads -> True]]]}]
$TMHOR$FreshVar[ l_List, ex_List, SEQ1$[ orig_Symbol]] :=
	VAR$[ {$varIndex++, SEQ1$[ $TMHOR$FreshName[ orig, Cases[ {l, ex}, (VAR$[ SEQ1$[ s_Symbol]]|VAR$[ {_, SEQ1$[ s_Symbol]}]) :> s, {0, Infinity}, Heads -> True]]]}]
$TMHOR$FreshVar[ l_List, ex_List, orig_Symbol] :=
	VAR$[ {$varIndex++, $TMHOR$FreshName[ orig, Cases[ {l, ex}, (VAR$[ s_Symbol]|VAR$[ {_, s_Symbol}]) :> s, {0, Infinity}, Heads -> True]]}]
	
$TMHOR$FreshName[ Null|$TMHOR$Unknown, l_List] := $TMHOR$FreshName[ Theorema`Knowledge`VAR$x$TM, l]
$TMHOR$FreshName[ x_Symbol, l_List] := freshName[ x, l]
	
$TMHOR$RangeToCondition[ rng:RNG$[ __]] :=
	With[ {cl = rngToCondition[ rng]},
		Switch[ cl,
			{___, $Failed, ___},
			$TMHOR$Unknown,
			{_},
			First[ cl],
			{},
			True,
			_,
			And$TM @@ cl
		]
	]
$TMHOR$RangeToCondition[ rng:Except[ _RNG$]] :=
	With[ {cl = Theorema`Language`FormulaManipulation`Private`singleRngToCondition[ rng]},
		Switch[ cl,
			{___, $Failed, ___},
			$TMHOR$Unknown,
			{_},
			First[ cl],
			{},
			True,
			_,
			And$TM @@ cl
		]
	]
	
(* 'substitute[ {pos1\[UpArrow], ...}, new, {{expr11, ..., expr1n, bound1\[UpArrow], old1}, ...}, trust ]'
first removes all '{expri1, ..., boundi, oldi}' where all the 'exprij' are '$TMHOR$Unknown'.
If no such list remains, it returns 'n' copies of '$TMHOR$Unknown', otherwise it searches for a list where 'new'
and 'oldi' are identical. In case it finds one, the function returns the sequence of the corresponding 'exprij',
otherwise it forgets all lists except the first one ('{expr1, ..., exprn, bound, old}') and then does the following:

- If none of the 'posi' is a list and 'trust' is 'True', then all positions of subexpressions of 'expr' matching
	any of the elements of 'old' are collected and gathered according to the element they match. Afterwards, the
	'posi' are set to those positions, and the function continues as if the 'pos' were lists from the very
	beginning.
- If all the 'posi' are lists of lists, 'substituteAt[ {pos1, ...}, expr, bound, new ]' is called.

If some of the 'posi' are lists but others are not, '$TMHOR$Unknown' is returned. *)
$TMHOR$Substitute[ pos_, new_, exprs:{___, {$TMHOR$Unknown..., _, _}, ___}, trust___] :=
	With[ {l = DeleteCases[ exprs, {$TMHOR$Unknown..., _, _}]},
		Switch[ l,
			{},
			$TMHOR$MakeSequence[ Table[ $TMHOR$Unknown, {Length[ First[ exprs]] - 2}]],
			_,
			$TMHOR$Substitute[ pos, new, l, trust]
		]
	]
$TMHOR$Substitute[ pos_, new_, {___, {expr___, _, new_}, ___}, ___] := expr
$TMHOR$Substitute[ pos:{{___List}..}, new_, {{expr___, bound_, old_}, ___}, ___] :=
	Module[ {nn, i = 0, del = {}},
		nn = DeleteCases[ MapThread[ (++i; Switch[ #2, #1|$TMHOR$Unknown, AppendTo[ del, {i}]; Null, _, #1])&,
										{new, old}
						],
						Null
			];
		Switch[ nn,
			{},
			(* We cannot just write 'expr' here, because otherwise the sequence would not be kept. *)
			$TMHOR$MakeSequence[ {expr}],
			_,
			$TMHOR$MakeSequence[ $TMHOR$SubstituteAt[ Delete[ pos, del], {expr}, bound, nn]]
		]
	]
$TMHOR$Substitute[ pos:{__Symbol}, new_List, {{expr___, bound_, old_List}, ___}, True] /; !MemberQ[ old, $TMHOR$Unknown] :=
	With[ {l = {expr}},
		pos = Map[ Position[ l, #]&, old];
		$TMHOR$MakeSequence[ $TMHOR$SubstituteAt[ pos, l, bound, new]]
	]
$TMHOR$Substitute[ ___] := $TMHOR$Unknown

$TMHOR$DisjointPositions[ {}] := True
$TMHOR$DisjointPositions[ {first_List, rest___List}] := $TMHOR$DisjointPositions[ first, {rest}, {rest}]
$TMHOR$DisjointPositions[ _, {}, {first_, rest___}] := $TMHOR$DisjointPositions[ first, {rest}, {rest}]
$TMHOR$DisjointPositions[ _, {}, {}] := True
$TMHOR$DisjointPositions[ p_, {q_, rest___}, r_] :=
	With[ {lp = Length[p], lq = Length[ q]},
		If[ (lp <= lq && $TMHOR$IsSubposition[ q, p]) || (lq < lp && $TMHOR$IsSubposition[ p, q]),
			False,
		(*else*)
			$TMHOR$DisjointPositions[ p, {rest}, r]
		]
	]
	
$TMHOR$IsSubposition[ {a___, ___}, {a___}] := True
$TMHOR$IsSubposition[ _, _] := False

(* '$TMHOR$SubstituteAt[ pos, expr, bound\[UpArrow], new ]' substitutes 'new' at all positions 'pos' in 'expr'.
It takes care of avoiding variable capture, meaning that it renames variables bound in 'expr' if necessary; For
this, it uses the information about variables bound in 'expr' stored in 'bound', which is supposed to be a list.
If 'bound' is no list but a symbol, this information is computed and 'bound' is set to the resulting list. *)
$TMHOR$SubstituteAt[ {{}...}, expr_List, _, _] := expr
$TMHOR$SubstituteAt[ pos_, expr_List, bound_Symbol, new_] :=
	$TMHOR$SubstituteAt[ pos, expr, bound = $TMHOR$CollectBound[ expr], new]
$TMHOR$SubstituteAt[ pos_, expr_List, bound_List, new_] :=
	Module[ {vars = DeleteDuplicates[ Cases[ new, (VAR$[ name:Except[ _List]]|VAR$[ {_, name_}]) :> name, {0, Infinity}, Heads -> True]],
			 rn, e, nn},
		rn = Cases[ bound, HoldPattern[ name_ -> {r_, p_}] /; (MemberQ[ vars, name] && $TMHOR$InScope[ pos, r]) :> {p, name}];
		e = Switch[ rn,
			{},
			expr,
			_,
			vars = Join[ vars, DeleteDuplicates[ Cases[ expr, (VAR$[ name:Except[ _List]]|VAR$[ {_, name_}]) :> name, {0, Infinity}, Heads -> True]]];
			ReplacePart[ expr,
				Replace[ rn, {p_, name_} :> (nn = $TMHOR$FreshName[ name, vars]; AppendTo[ vars, nn]; p -> VAR$[ {$varIndex++, nn}]), {1}]
			]
		];
		ReplacePart[ e, MapThread[ Rule, {pos, new}]]
	]
	
$TMHOR$InScope[ {p___List}, r_List] := Apply[ Or, Map[ $TMHOR$InScope[ #, r]&, Hold[ p]]]
$TMHOR$InScope[ {a___, __}, {a___}] := True
$TMHOR$InScope[ {a___, x_, ___}, {a___, y_}] := (x > y)
$TMHOR$InScope[ {a___, x_, ___}, {a___, y_, _}] := (x>y)
$TMHOR$InScope[ _List, _List] := False

(* 'collectBound[ expr ]' returns a list '{name1 -> {r1,p1}, ...}' of all variables that are bound in 'expr'
(together with a list of their occurrences).
Note that 'ri' refers to the position of the range (SIMPRNG$ etc.) where 'vari' is bound, whereas 'pi' is a list
of _all_ occurrences (including the range) of 'vari' in 'expr'.
The 'namei' are the _base names_ of the variables, hence it is possible that several 'namei -> {ri,pi}' have the
same 'namei'. *)
$TMHOR$CollectBound[ expr_] :=
	Module[ {sub, name},
		Flatten[ Map[ Function[ p,
							sub = Switch[ p, {}, expr, _, Extract[ expr, p]];
							MapIndexed[ Function[ {v, i},
											name = Replace[ v, {VAR$[ {_, n_}] :> n, VAR$[ n:Except[ _List]] :> n}];
											name -> {Join[ p, {1, First[ i]}], Map[ Join[ p, #]&, Position[ sub, v]]}
										],
										rngVariables[ First[ sub]]
							]
					],
					Position[ expr, _[ _RNG$, ___]]
				],
				1
			]
	]
	
$TMHOR$Extract[ $TMHOR$Unknown, l:{___List}] := Table[ $TMHOR$Unknown, {Length[ l]}]
$TMHOR$Extract[ expr_, l:{___List}] := Map[ Switch[ #, {}, expr, _, Extract[ expr, #]]&, l]

(* 'makeInstance[ inst, pspec, args, loc, expr]' heavily relies on the fact that any two distinct variables in
'inst' and 'expr' have different names, which is ensured by 'preprocessExpression'. *)
$TMHOR$MakeInstance[ inst:($TMHOR$Lambda[ rng_RNG$, $TMHOR$Unknown]), args_List, loc_List, expr:Except[ $TMHOR$Unknown]] :=
	Module[ {i = 0, newRng = rng, repl, nv, allvars},
		allvars =
			DeleteDuplicates[
				Cases[ {rng, expr}, v_VAR$ :> SymbolName[ Theorema`Language`FormulaManipulation`Private`varName[ v]], {0, Infinity}]
			];
		repl =
			DeleteCases[
				MapThread[
					Function[ {a, v, l},
						++i;
						If[ l === {},
							Null,
							With[ {name = Theorema`Language`FormulaManipulation`Private`varName[ v]},
								nv = freshName[ SymbolName[ name], allvars];
								AppendTo[ allvars, nv];
								nv = VAR$[ {++$varIndex, Symbol[ Context[ name] <> nv]}];
								If[ nv =!= v, newRng = newRng /. v -> nv];
								a -> nv
							]
						]
					],
					{args, rngVariables[ rng], loc}
				],
				Null
			];
   		$TMHOR$MakeInstance[ $TMHOR$Lambda[ newRng, expr /. repl]]
   	];
$TMHOR$MakeInstance[ $TMHOR$Lambda[ rng_RNG$, SEQ$[ inner_]], ___] :=
	$TMHOR$Lambda[ rng, inner]
$TMHOR$MakeInstance[ $TMHOR$Lambda[ rng_RNG$, inner:SEQ$[ ___]], ___] :=
	Map[ $TMHOR$Lambda[ rng, #]&, inner]
$TMHOR$MakeInstance[ SEQ$[ a_], ___] :=
	a
$TMHOR$MakeInstance[ inst_, ___] := inst

End[] (* End Private Context *)
EndPackage[]
